\ Theseus manual

\ ==========
\ *! theseus
\ *T Theseus GUI designer
\ ==========

\ **********************
\ *S Introducing Theseus
\ **********************
\ *P Minos and Theseus are the answer to the question "Is there
\ ** something like Visual BASIC or Delphi in Forth?". Basically,
\ ** these GUI systems contain two components: a library with a
\ ** wide variety of elements for drawing a graphical user interface;
\ ** e.g. windows, buttons, edit-controls, drawing areas, etc.
\ ** and an editor to select and combine the elements with the mouse by
\ ** drag-and-drop or click-and-point actions. You then insert
\ ** code that acts when buttons are pressed.

 \ *P Minos is the drawing library, Theseus is the GUI editor.
 \ ** When you save your design Theseus generates the Forth
 \ ** source code to do what you specified. You create your
 \ ** application by including Minos and the generated code in
 \ ** your application.

\ *P It isn't really necessary to paint graphical user interfaces
\ ** together, in the same way as it isn't for text, e.g. TeX versus
\ ** Open Office. Many typesetting functions are more semantic than
\ ** visual, e.g. a piece of text is a headline or emphasized instead
\ ** of written in bold 18 point Garamond or 11 point Roman italic.
\ ** All this is true for user interface design too, and to some extent
\ ** much more. It's not the programmer who decides which font and
\ ** size to use for the User Interface - that's up to the user, as
\ ** are the colour of buttons and texts.

\ *P To lay out individual widgets, more abstraction than defining
\ ** position, width and height makes sense. Typically buttons are
\ ** arranged horizontally or vertically, perhaps with some space
\ ** between them. The size of buttons must follow the containing
\ ** strings, and should conform to a consistent aesthetic, e.g.
\ ** each button in a row has the same width.

\ *P Such an abstract model, related to TeX's boxes and glues model,
\ ** performs quite well even without a visual editor. The programmer
\ ** isn't responsible for typesetting the buttons and boxes.
\ ** This approach is quite usual in Unix. Motif and Tcl/Tk use
\ ** neighborhood relations, Interviews uses boxes and glues. I
\ ** decided for boxes and glues, since it's a fast and intuitive
\ ** solution, although it needs more objects to get the same result.
\ ** Boxes form the components. The space between them is defined by
\ ** glues.

\ *P These concepts contradict somehow with a graphical editing process,
\ ** since the editors I know don't provide abstract concepts such as
\ ** "place left of an object" or place in a row, but positions.

\ *P Theseus works with a relationship model. It does not work in
\ ** terms of the x,y,w,h positioning model familiar to Windows
\ ** programmers. Instead you specify the relationship between
\ ** items in Theseus, and then Minos displays the items.

\ *P Theseus is designed to create Minos dialogs. Each dialog is
\ ** derived from a window class. You can create dialogs by composing
\ ** boxes (layout managers) and widgets together, add actions to
\ ** the widgets, add variables and methods to the derived window
\ ** classes and so on.

\ *P Minos has four sets of classes:
\ *(
\ *B *\b{Widgets}: "window gadgets", these are the basic objects like
\ ** buttons, labels, icons and text fields.
\ *B *\b{Displays}: these widgets are the drawing area of other
\ ** widgets. They know how to draw to X or to the parent display.
\ ** It is also possible to create new displays that, e.g., draw
\ ** into an OpenGL widget or send drawing events over the network,
\ ** to copy drawing messages to more than one parent display
\ ** and so on.
\ *B *\b{Boxes}: or layout managers. Although each object in Minos has
\ ** coordinates and size, formating is done by a surrounding
\ ** layout manager. These either form horizontal boxes, and
\ ** align their contents from left to right, or vertical boxes,
\ ** and align their contents from top to bottom. Each widget
\ ** has a horizontal and a vertical glue, which are used to
\ ** compute the look of the widgets.
\ *B *\b{Actions}: these are the objects (Forth words) that link data and
\ ** widgets together. An action knows in which state it is,
\ ** and what to do when the state changes. There are several
\ ** types of actions, for simple buttons, toggle buttons,
\ ** text fields, and sliders.
\ *)

\ *P Each dialog is hierarchically composed from widgets,
\ ** boxes, and displays (e.g. viewports). Each active object,
\ ** thus each button, toggle button, slider, and text field
\ ** has an associated action, a name, and other attributes.


\ ******************************
\ *S Compiling Minos and Theseus
\ ******************************
\ *P If you have not been supplied with a precompiled version
\ ** of Theseus or you have upgraded the Forth kernel, you will have to
\ ** rebuild Theseus. The incantations are specific to the host
\ ** Forth. To run Theseus as it is is designed to be run, you
\ ** should create an executable with a precompiled version
\ ** of Minos. Note that this may modify the start-up behaviour
\ ** and command-line processing of your Forth.

\ *P For bigForth, build *\i{xbigforth}:
\ *E bigforth
\ **  include startx.fs
\ *P For VFX Forth, build *\i{xvfxlin}:
\ *E vfxlin
\ **  include startx.fs
\ **  save xvfxlin

\ *P You can now run the binary, which will appear in a window.
\ ** We usually refer to this binary as *\i{<xbin>}.

\ ====================
\ *S Launching Theseus
\ ====================
\ *P In most cases, Theseus is compiled from source code every time
\ ** - it only takes a fraction of a second. Theseus will start
\ ** when the code has compiled.
\ *C   <xbin> theseus.fs

\ *P This will leave you with the following display.
\ *R @* @center @image{Images/minostheseus,6in,,,.png}

\ *P You start a new design using the top menu bar.
\ *R @* @center @image{Images/menu,,,,.png}
\ *P The file menu allows you to load and save designs. The edit
\ ** menu has only two entries, for a new dialog and a new menu
\ ** window.

\ *P Once you have selected a menu or dialog to buid, you can then
\ ** select what you are going to put in it using the widget bar
\ ** below the menu.
\ *R @* @center @image{Images/widget,,,,.png}
\ *P The widget bar contains groups of widgets. To insert a
\ ** widget into a dialog, select the group and click on a button.
\ ** The corresponding widget is inserted in the current position
\ ** (affected by the mode switches).

\ *P Rather than go into tedious detail about every button, we are
\ ** now going to take you step by step through the process of
\ ** building a simple calculator. After that, we'll build a menu
\ ** system.


\ *********************
\ *S Calculator example
\ *********************
\ *R @* @center @image{Images/widget,,,,.png}
\ *P The picture above shows what we are going to build. There's
\ ** a canned example *\i{calc.m} if you want to cheat. It really
\ ** is worth going through this example yourself.

\ *P Formatting buttons and text fields is done by the system, and
\ ** is thus not the task of the programmer (you), who only has to
\ ** specify the logical arrangement.

\ *P The project is hierarchically arranged. The topmost hierarchy
\ ** are the dialog windows. These windows understand two additional
\ ** methods, open and modal-open which allows you to create both
\ ** non-modal and modal dialogs. The user then creates a
\ ** framework of horizontal and vertical boxes inside the
\ ** dialog. These boxes are filled with contents and glues.
\ ** Glues are what we call the space between items. Glue can
\ ** be modified (by you) too.

\ *P Launch Theseus.
\ ** Create a "New Dialog" with the "Edit" menu. This is the
\ ** dialog bar at the start of the project:
\ *R @* @center @image{Images/step1,,,,.png}
\ *P Each object has a title and a name. The title is usually
\ ** displayed at the top of the object as a caption. If the
\ ** title field shows "No Title", there will be no caption.
\ ** The name is the name of the Forth object that will be
\ ** generated when Theseus generates code for you. For the
\ ** moment, we will not bother much with titles or names, but
\ ** they will be important when we get to the menu example.

\ *P Next we need three entries that contain a label and an
\ ** editable number. Click in the box in the new dialog that
\ ** has a cross in it. The cross indicates an empty area.
\ ** Select "Text Fields" in the widget bar, and then click
\ ** "Infonumber" three times. You should see:
\ *R @center @image{Images/step2,,,,.png}
\ *P When you select each field in turn, you will see that the
\ ** box inspector has replaced the Minos logo. Edit three of
\ ** the four fields
\ *(
\ *B *\b{Infotextfield name} - this is the name (in Forth) of
\ ** the object. We need the name in order to get at the data
\ ** in the number field. Call these *\fo{a#} and *\fo{b#} for
\ ** the inputs, and *\fo{r#} for the result.
\ *B *\b{Number} - the initial value that will appear in the
\ ** edit box.
\ *B *\b{string} - this is the label at the left of each entry.
\ ** Suitable strings are *\i{A:}, *\i{B:} and *\i{R:}.
\ *)

\ *P Beneath the two input fields the operation buttons should
\ ** be arranged one aside each other. A horizontal box (hbox)
\ ** does the job, with four buttons in it.
\ *( 1
\ *B First click to select the second input box, the one labelled
\ ** B.
\ *B At the left of  the main Theseus window, hover the mouse over the top four
\ ** icons in the vertical bar. Tooltips will appear to show
\ ** their action. Select the one whose tip says "Add after
\ ** current object".
\ *B Select hbox (near upper right) to make the box.
\ *B Click the "Add last in box" icon.
\ *B Click into the new box and click off the *\b{vskip} button
\ ** on the right hand side of the main Theseus window.
\ *B Click "Buttons" on the widget bar, and then click "Button"
\ ** on the next line four times to put the buttons in the hbox.
\ *B Select each button in turn and label it for one of the
\ ** primary operations +, -, *, /. After selecting it, you
\ ** can press <Esc> to clear the text field first and then just
\ ** type the text.
\ *)
\ *P Now you should see something like this:
\ *R @* @center @image{Images/step3,,,,.png}

\ *P You have already given object names to the input and result
\ ** boxes. Now we must add the action code in the *\b{Code:}
\ ** field of each button. The code looks as follows, for
\ ** +, -, *, and /.
\ *E a# get  b# get  d+  r# assign
\ ** a# get  b# get  d-  r# assign
\ ** a# get  b# get  d*  r# assign
\ ** a# get  b# get  drop ud/mod  r# assign  drop
\ *P We will discuss the gory details of accessing object data
\ ** later. For the moment, we'll just note that *\fo{get} and
\ ** *\fo{assign} are methods applied to the objects *\fo{a#},
\ ** *\fo{b#} and *\fo{r#}. You should have a button inspector
\ ** display like the one below.
\ *R @center @image{Images/step4,,,,.png}

\ *P It may be useful to take the result and copy it to one of
\ ** the input buttons for reuse. Thus two additional buttons
\ ** are required, and to make it nice, all buttons should have
\ ** the same size (with "tabbing" box style). The dialog should
\ ** have a title, and a name, too. The added code for the *\fo{>A}
\ ** and *\fo{>B} buttons is:
\ *E r# get  a# assign
\ ** r# get  b# assign
\ *P Now you should see something like this:
\ *R @* @center @image{Images/step5,,,,.png}

\ *P Save the application. You must provide the extension
\ ** for the file name. By convention, Theseus uses a *\i{.m}
\ ** extension for its auto-generated files.

\ *P If your particular host Forth permits it, you can click
\ ** the *\b{Run application} button and make your calculator
\ ** do some calculations.


\ *********************
\ *S Concepts for menus
\ *********************
\ *P Remember that Minos builds things from boxes and glues. A
\ ** menu is just a collection of labelled boxes that do things
\ ** when pressed. In Minos terms, a top-level menu entry that
\ ** just runs another menu is a *\b{menu-title}. An entry that
\ ** actually does something is a *\b{menu-entry}. A menu entry
\ ** (not at the top level) that launches another menu is a
\ ** *\fo{sub-menu}.

\ *P A top-level menu is built either as a menu bar using an
\ ** dialog, or as a "New menu window", which is
\ ** an *\fo{box} containing menu items and a panel for the
\ ** application's display. All other menus are built with
\ ** dialogs containg menu items.

\ *P To link them together, the *\b{menu-title}s point to the dialogs
\ ** containg menu items and so on. Note that initially you only want to
\ ** see the top-level menu, so all the dialogs for sub-menus have
\ ** their "Show dialog" buttons (see the left-hand buttons on
\ ** each dialog editor) set to "not shown" (red cross rather than
\ ** green mark). This prevents all the menus being visible at
\ ** application start-up.


\ ***************
\ *S Menu example
\ ***************
\ *P The example here was generated during the port of Minos and
\ ** Theseus to VFX Forth for Linux.
\ *R @* @center @image{Images/vfxmenu,,,,.png}

\ =================
\ *N Top level menu
\ =================
\ *P Start Theseus and then
\ *( 1
\ *B Select *\b{Add last in box} (the default) and
\ *B Add a *\b{New Menu Window} from the *\b{Edit} menu.
\ *B Select the box portion at the top, and click the *\b{border}
\ ** button to make the menu more distinct.
\ *B In the dialog bar (a menu window is a special case of a
\ ** dialog), give the dialog a title (caption), and a name for
\ ** the Forth object. We used *\fo{MainMenu}, because that's
\ ** what we were prototyping then.
\ *)
\ *P You will see a horizontal box at the top and an empty panel
\ ** below it. The box will contain the top-level menu, and the
\ ** panel will contain the other parts of the application. We
\ ** are now ready to build the top-level menu. The picture below
\ ** shows the Theseus display midway through the process.
\ *( 1
\ *B Select *\b{Menus} on the widget bar, and add three *\b{Menu-Title}s.
\ *B Select *\b{Glues} on the widget bar, and add one *\b{HGlue}
\ *B Go back to *\b{Menus} and add another *\b{Menu-Title}.
\ *B Select each menu entry in turn and edit the *\b{Menu-Title name}
\ ** and *\b{String} fields. The names are the names of Forth objects.
\ ** Since this is the main menu and the first entry is for a File
\ ** menu, we name it *\fo{mmFile} and enter *\b{File} as the string.
\ ** We will edit the *\b{Menu} field when we have built the File
\ ** menu itself.
\ *B Select the *\b{HGlue} by selecting the gap between the third
\ ** and fourth menu titles. Check the *\b{fill} button to keep
\ ** the *\b{Help} menu at the right. If you don't like this
\ ** layout you can remove the *\b{HGlue} by selecting it with
\ ** a shift-left-mouse-button click.
\ *)
\ *R @* @center @image{Images/menuexample1,,,,.png}

\ ============================
\ *N Adding second level menus
\ ============================
\ *P Next, we need to add the four menus. These are vertically
\ ** oriented dialogs containing *\b{Menu-Entry} components.
\ ** The process is as follows.
\ *( 1
\ *B Create a new dialog. It should not have a title, but it
\ ** needs a name. Again this is a Forth object. For the File
\ ** menu, we called it *\fo{FileMenu}.
\ *B Give the dialog a border and change the *\b{Show Dialog}
\ ** button on the dialog bar to a red cross. This stops the
\ ** dialog being displayed until you actually select it.
\ *B Add your *\b{Menu-Entry} items. Each one needs a Forth
\ ** name in the *\b{Menu-Entry name} field and some text in
\ ** the *\b{String} field.
\ *B To make the menu entry do something when selected,
\ ** you can put a code fragment in the *\fo{Code} field.
\ ** For the moment leave this field blank. See the section
\ ** "Adding actions" below for more details.
\ *)

\ =======================
\ *N Reviewing the design
\ =======================
\ *P Click the *\b{Run application} button and press the
\ ** menu entries. The menus appear as they will appear in your
\ ** final application.

\ *P At first glance the boxes are too deep - there's far too
\ ** much white space, and the items could be grouped by the
\ ** use of separators. We'll do this in the next section.

\ *P After reviewing a design, remember to close the test design
\ ** otherwise you'll find the screen cluttered with examples!


\ ===================
\ *N Making it pretty
\ ===================


\ ********************************
\ *S Adding a menu without a panel
\ ********************************


\ *********************
\ *S Navigating Theseus
\ *********************

\ =============
\ *N At the top
\ =============
\ *R @* @center @image{Images/menu,,,,.png}
\ *P The menu bar contains a file menu, an edit menu and a help
\ ** menu. From the *\b{Edit} menu, you select a new dialog or
\ ** menu.

\ *R @* @center @image{Images/widget,,,,.png}
\ *P The widget bar contains groups of widgets. To insert a widget into
\ ** a dialog, select the group and click on a button. The corresponding
\ ** widget is inserted in the current position (affected by the mode
\ ** switches).

\ ================
\ *N Editing modes
\ ================
\ *R @* @center @image{Images/modes,,,,.png}
\ *P The editing modes are separated into three groups:
\ *( 1
\ *B Insert modes:
\ *(
\ *B Add object first in current box
\ *B Add object last in current box
\ *B Add object before current box
\ *B Add object after current box
\ *)
\ *B Navigation through nested boxes, sets the active box.
\ ** The reason why there is a navigator is that sometimes boxes
\ ** are unselectable because they have no border and the inside
\ ** boxes fill them completely, leaving no outer space to select
\ ** with a mouse click.
\ *(
\ *B Move one box up in hierarchy
\ *B Move one box to the left/up
\ *B Move one box to the right/down
\ *B Move to the first child
\ *)
\ *B Short cuts:
\ *(
\ *B  Load dialog
\ *B  Save dialog
\ *B  Try dialog
\ *B  Save as module
\ *)
\ *)

\ ================
\ *N Dialog editor
\ ================
\ *R @* @center @image{Images/dialog,,,,.png}
\ *P The dialog editor shows a navigation bar for each dialog to
\ ** edit. The dialog itself is resizable with the split bar below and on
\ ** the side, to see how it looks resized.

\ *P The navigation bar consists of an icon to open/hide the dialog, to
\ ** open/hide the declarations field, to open/hide the code field, to
\ ** select whether to show the dialog, a dialog menu, a dialog title and a
\ ** dialog name. It is important to give every dialog a name, since
\ ** dialogs without names can't be saved. The dialog name is the name of
\ ** the derived class, you can refer to this class in your code.

\ *P The declaration field contains variable and method declarations of
\ ** the dialog class. The code field contains method definitions.
\ ** The dialog edit field contains the dialog itself.

\ *P Use the mouse for selection.
\ *(
\ *B [Edit] A simple mouse click opens the inspector of that object, with
\ ** focus on the text/code/name field (left/middle/right mouse button).
\ *B [Cut&Paste] Shift-mouse} clicking left cuts the object to the cut stack,
\ ** clicking middle or right pastes from the cut stack.
\ *B [Try] Cintrol-mouse clicking makes the object react as in the dialog,
\ ** but without executing code.
\ *)

\ *P You can delete a complete dialog using *\b{Dialog Menu > Cut Dialog}
\ ** in the dialog's navigation bar.

\ *P When you are happy with a dialog, you can fold it to consume
\ ** less Theseus display space using the *\b{Dialog Editor} button.
\ ** Clicking this button again will unfold (restore) the view.


\ ==============
\ *N Box creator
\ ==============
\ *R @* @center @image{Images/newbox,,,,.png}
\ *P The box creator has two buttons to create horizontal and vertical
\ ** boxes. Boxes are simple layout managers, that arrange containing
\ ** objects one after the other. Boxes are created as normal objects, so
\ ** they go to the same places where a normal object would go. They
\ ** inherit the settings of the parent object, so these settings have to
\ ** be changed using the box inspector.

\ ================
\ *N Box inspector
\ ================
\ *R @* @center @image{Images/box,,,,.png}
\ *P The properties of the current box can be changed in the box
\ ** inspector on the right hand side of the Theseus screen:
\ *(
\ *B  The *\b{horizontal} switch changes the direction of the box
\ *B  The *\b{activate} switch changes the selection behavior: active
\ ** boxes contain one single active object, navigation with
\ ** <tab> is possible.
\ *B  The *\b{radio} switch enables deselection on click, thus only
\ ** one switch inside such a box may be active at a time.
\ *B  The *\b{tabbing} switch changes the layout: all objects except
\ ** glues in tabbed boxes have the same size.
\ *B  The *\b{hfixbox} shrinks the box to the minimal size, no growing
\ **is possible in horizontal direction
\ *B  The *\b{vfixbox} shrinks the box to the minimal size, no growing
\ ** is possible in vertical direction
\ *B  The *\b{flipbox} hides the box when active
\ *B  The *\b{hskip} box or slider (with *\b{Details} activated)
\ ** adds horizontal skips between objects
\ *B  The *\b{vskip} box or slider (with *\b{Details} activated)
\ ** adds vertical skips between objects
\ *B  The *\b{border} box or slider (with *\b{Details} activated)
\ ** adds a shadow to the box (raised or sunken).
\ *)


\ ===================
\ *N Object inspector
\ ===================
\ *R @* @center @image{Images/inspector,,,,.png}
\ *P The object inspector contains the informations of the current
\ ** object. The fields depend on the class of the object, however, some
\ ** fields are common between objects.
\ *(
\ *B  The *\b{name} field selects the name of the object, this
\ ** name is used in code to refer to this object
\ *B  The *\b{string} field is the string the object displays
\ *B  The *\b{code} field is the code that is executed on clicks
\ *B  The *\b{tooltip} field is the tooltip that is shown when the
\ ** mouse is over the object (an empty string means no tooltip)
\ )
\ *P There are other fields for properties specific to each type
\ ** of widget.


\ *****************
\ *S Adding actions
\ *****************


\ ======
\ *> ###
\ ======
