\ Doubles.fth - VFX Forth double and quad number support

((
Copyright (c) 1990..2004, 2006, 2008, 2012, 2013
Construction Computer Software
Cape Town
South Africa

Copyright (c) 2005, 2006, 2012, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

DOUBLES.FTH contains double, triple and quadruple precision
arithmetic operations.

No word may be removed or have changed effects without permission
from CCS.


To do
=====

Change history
==============
20130924 MPE007 Corrected UTT/MOD.
20130520 CCS006 Added (TT+).
		Added and corrected QN/MOD and UQN/MOD.
		Added more CCSisms.
20120917 CCS005 Added T@ T! Q@ Q!.
		Renamed T<<1 to T2*.
		Added 3SWAP and 3OVER.
20120512 MPE004 Added UD/MOD and UDS/MOD
20120430 MPE003 DocGen and code tidy.
20060816 MPE002 Overhauled to separate generic and CCS extensions.
20041201 MPE001 Modified to use ANS Forth naming conventions.
                Optimised for VFX code generators.
))

decimal

\ ==========
\ *! doubles
\ *T Double, Triple and Quad number support
\ ==========
\ *P The file *\i{Lib\Doubles.fth} contains double, triple and
\ ** quadruple precision arithmetic operations. These words are
\ ** in regular use by *\i{Construction Computer Software} (CCS).
\ ** No word will be removed or changed without permission from
\ ** CCS.

[undefined] CCSisms? [if]
1 constant CCSisms?	\ -- 0/1
\ *G Change to zero not to compile CCS specific code. This
\ ** version is used only if the word has not already been
\ ** defined. This allows you to define *\fo{CCSisms?} as you
\ ** want before compiling this file.
[then]


\ ***********
\ *S Notation
\ ***********
\ *P The following stack notation is used:
\ *D d  double (64 bit) signed value.
\ *D t  triple (96 bit) signed value.
\ *D b  d.dpl, double (64 bit) signed value and single (8 bit) signed dpl.
\ *D q  quad (128 bit) signed value.
\ *D s  s.dpl single (32 bit) signed value and single (8 bit) signed dpl.
\ *D n  single (32 bit) signed value.
\ *D u  unsigned single (32 bit) unsigned value.
\ *D c  character (8 bit) unsigned value.
\ *D w  word (16 bit) unsigned value.
\ *D r  real (floating point) on floating point stack
\ *D f  real (floating point) on floating point stack
\ *D d.pl double (64 bit) signed value and single (8 bit) signed dpl.
\ *D x.dpl first d.dpl parameter.
\ *D y.dpl 2nd d.dpl parameter.
\ *D z.dpl result d.dpl parameter.

\ *********************
\ *S Storage Byte Sizes
\ *********************
\ *P The following notation is used for storage sizes:
\ *D kilobyte  (kB/KB) 10^3
\ *D megabyte  (MB)    10^6
\ *D gigabyte  (GB)    10^9
\ *D terabyte  (TB)    10^12
\ *D petabyte  (PB)    10^15
\ *D exabyte   (EB)    10^18
\ *D zettabyte (ZB)    10^21
\ *D yottabyte (YB)    10^24

\ ****************
\ *S Double values
\ ****************
\ *P The Maximum values for doubles:
\ *D Quintillion  1,000,000,000,000,000,000      10^18
\ *D Quadrillion      1,000,000,000,000,000      10^15
\ *D Trillion             1,000,000,000,000      10^12
\ *D Billion/Milliard         1,000,000,000      10^9
\ *D Million                      1,000,000      10^6
\ *D Lakh                           100,000      10^5
\ *D Myriad                          10,000      10^4
\ *D Zero dpl     9,223,372,036,854,775,807
\ *D Two  dpl        92,233,720,368,547,758.07
\ *D Four dpl           922,337,203,685,477.5807
\ *D Six  dpl             9,223,372,036,854.775807
\ *D Seven dpl              922,337,203,685.4775807
\ *D Eight dpl               92,233,720,368.54775807

\ *S Extreme constants
  $80000000 CONSTANT NaN
\ *G 32bit no value
: D-NaN                                                       ( --- 0 nan )
   \ *G 64bit NaN
   $00000000 Nan
  ;

: D.DPL-NaN                                                       ( --- 0 nan )
   \ *G 64bit NaN and DPL
   D-NaN #128 ;

: D.DPL-NaN?                                                 ( d.dpl --- bool )
   DROP D-NaN D= ;

: D.DPL?                                                       ( d.dpl -- flag )
   \ *G True if d.dpl has a value
   DROP D0<> ;

\ *****************
\ *S Coded routines
\ *****************

code t+		\ t1l t1m t1h t2l t2m t2h -- t3
\ *G Add two triple precision (96 bit) integers.
                \ 16  12  8   4   0   EBX
   mov   eax, #16 [ebp]			\ low items
   add   eax, 4 [ebp]
   mov   #16 [ebp], eax
   mov   eax, #12 [ebp]			\ middle items
   adc   eax, 0 [ebp]
   mov   #12 [ebp], eax
   adc   ebx, 8 [ebp]			\ high items
   lea   ebp, #12 [ebp]			\ update stack
   next,
end-code

code t-		\ t1l t1m t1h t2l t2m t2h -- t1-t2
\ *G Subtract two triple precision (96 bit) integers.
                \ 16  12  8   4   0   EBX
   mov   eax, #16 [ebp]			\ low items
   sub   eax, 4 [ebp]
   mov   #16 [ebp], eax
   mov   eax, #12 [ebp]			\ middle items
   sbb   eax, 0 [ebp]
   mov   #12 [ebp], eax
   mov   eax, 8 [ebp]			\ high items
   sbb   eax, ebx
   mov   ebx, eax
   lea   ebp, #12 [ebp]			\ update stack
   next,
end-code

code tnegate	\ t1 -- t2
\ *G Negate triple precision (96 bit) integer.
  not   4 [ebp]
  not   0 [ebp]
  not   ebx
  xor   edx, edx
  add   4 [ebp], # 1
  adc   0 [ebp], edx
  adc   ebx, edx
  next,
end-code

code t2*	\ t -- t<<1
\ *G Shift triple number 1 bit to left.
  shl   4 [ebp], # 1
  rcl   0 [ebp], # 1
  rcl   ebx, # 1
  next,
end-code

code q+		\ q10 q11 q12 q13 q20 q21 q22 q23 --
\ *G Add two quadruple precision (128 bit) integers.
                \ 24  20  16  12  8   4   0   EBX
   mov   eax, #24 [ebp]			\ lowest items
   add   eax, 8 [ebp]
   mov   #24 [ebp], eax
   mov   eax, #20 [ebp]			\ ls middle items
   adc   eax, 4 [ebp]
   mov   #20 [ebp], eax
   mov   eax, #16 [ebp]			\ ms middle items
   adc   eax, 0 [ebp]
   mov   #16 [ebp], eax
   adc   ebx, #12 [ebp]			\ high items
   lea   ebp, #16 [ebp]			\ update stack
   next,
end-code

code q-		\ q10 q11 q12 q13 q20 q21 q22 q23 -- q1-q2
\ *G Subtract two quadruple precision (128 bit) integers.
                \ 24  20  16  12  8   4   0   EBX
   mov   eax, #24 [ebp]			\ low items
   sub   eax, 8 [ebp]
   mov   #24 [ebp], eax
   mov   eax, #20 [ebp]			\ ls middle items
   sbb   eax, 4 [ebp]
   mov   #20 [ebp], eax
   mov   eax, #16 [ebp]			\ ms middle items
   sbb   eax, 0 [ebp]
   mov   #16 [ebp], eax
   mov   eax, #12 [ebp]			\ high items
   sbb   eax, ebx
   mov   ebx, eax
   lea   ebp, #16 [ebp]			\ update stack
   next,
end-code

code qnegate	\ q1 -- q2
\ *G Negate quadruple precision (128 bit) integer.
  not   8 [ebp]
  not   4 [ebp]
  not   0 [ebp]
  not   ebx
  xor   edx, edx
  add   8 [ebp], # 1
  adc   4 [ebp], edx
  adc   0 [ebp], edx
  adc   ebx, edx
  next,
end-code

code q2*	\ q -- q<<1
\ *G Shift quad number 1 bit to left.
  shl   8 [ebp], # 1
  rcl   4 [ebp], # 1
  rcl   0 [ebp], # 1
  rcl   ebx, # 1
  next,
end-code

: 3SWAP                                 \ x1 x2 x3 x4 x5 x6 -- x4 x5 x6 x1 x2 x3
   >R >R SWAP 2SWAP                                             \ 4 3 1 2 -- 6 5
   R> SWAP 2SWAP                                                \ 4 5 2 3 1 -- 6
   R> SWAP 2SWAP ;                                                 \ 4 5 6 1 2 3

: 3OVER                        \ x1 x2 x3 x4 x5 x6 -- x1 x2 x3 x4 x5 x6 x1 x2 x3
   5 PICK 5 PICK 5 PICK ;

\ ********************
\ *S Memory operations
\ ********************

: 3@		\ addr -- t
\ *G Fetch a triple. The most significant word is at the low address.
  dup 2 cells + @ swap			\ -- t.lo addr
  dup 1 cells + @ swap			\ -- t.lo t.mid addr
  @
;

: 3!		\ t addr --
\ *G store a triple. The most significant word is at the low address.
  tuck !  tuck cell + !  2 cells + !
;

: 4@		\ addr -- q
\ *G Fetch a quad. The most significant word is at the low address.
  dup 3 cells + @ swap			\ -- q.lo addr
  dup 2 cells + @ swap			\ -- q.lo q.ml addr
  dup 1 cells + @ swap			\ -- q.lo q.ml q.mh addr
  @					\ -- q
;

: 4!		\ q addr --
\ *G Store a quad. The most significant word is at the low address.
  tuck !  tuck cell + !  tuck 2 cells + !  3 cells + !
;


: D@		\ addr -- d
\ *G Fetch a double in native form. For the x86 ISA, the most
\ ** significant word is at the high address.
  dup @ swap cell + @
;

: D!		\ d addr --
\ *G store a double in native form. The most significant word is
\ ** at the high address.
  tuck cell + ! !
;

: T@		\ addr -- t
\ *G Fetch a triple in native form. For the x86 ISA, the most
\ ** significant word is at the high address.
  dup @ swap				\ -- t.lo addr
  dup cell + @ swap			\ -- t.lo t.mid addr
  2 cells + @
;

: T!		\ t addr --
\ *G store a triple in native form. The most significant word is
\ ** at the high address.
  tuck 2 cells + !  tuck cell + !  !
;

: Q@		\ addr -- q
\ *G Fetch a quad in native form. The most significant word is at
\ ** the high address.
  dup           @ swap			\ -- q.lo addr
  dup 1 cells + @ swap			\ -- q.lo q.ml addr
  dup 2 cells + @ swap			\ -- q.lo q.ml q.mh addr
      3 cells + @			\ -- q
;

: Q!		\ q addr --
\ *G Store a quad in native form. The most significant word is at
\ ** the high address.
  tuck 3 cells + !  tuck 2 cells + !  tuck 1 cells + !  !
;


\ ************************
\ *S High level primitives
\ ************************

: UP-PRECISION                                                  ( n1 --- n1 n2 )
   \ *G Increase the precision of n1 with the data stack width
   DUP 0< ;

[UNDEFINED] DU<	[IF]
: DU<		\ ud1 ud2 -- flag
\ *G True if ud1<ud2.
  ROT SWAP 2DUP U< IF
    4DROP True
  ELSE
    <>
    IF  2DROP False  ELSE  U<  THEN
  THEN
;
[THEN]

: TU<		\ ut1 ut2 -- flag
\ *G True if ut1<ut2.
  >r rot r>  2dup u< if			\ -- ud1 ud2 uh1 uh2
    2drop 2drop 2drop  true  exit
  endif
  <> if					\ -- ud1 ud2
    2drop 2drop  false  exit
  endif
  du<
;

: QU<		\ uq1 uq2 -- flag
\ *G True if uq1<uq2
  2>r rot 2r> swap -rot  2dup u< if	\ -- ut1 ut2 uh1 uh2
    2drop 2drop 2drop 2drop true  exit
  endif
  <> if					\ -- ut1 ut2
    2drop 2drop 2drop  false  exit
  endif
  du<
;

: tabs		\ t -- |t|
\ *G Return the absolute value of a triple.
  dup 0<
  if  tnegate  endif
;

: qabs		\ q -- |q|
\ *G Return the absolute value of a quad.
  dup 0<
  if  qnegate  endif
;

SYNONYM N>D UP-PRECISION   \ n --- d
\ *G Convert signed single value to signed double.

SYNONYM D>T UP-PRECISION   \ d -- t
\ *G Convert signed double to signed triple.

SYNONYM T>Q UP-PRECISION   \ t -- q
\ *G Convert signed triple to signed quad.

: D>Q		\ d -- q
\ *G Convert signed double to signed quad.
  UP-PRECISION dup
;

: N-SIGN	\ n -- |n| flag
\ *G Checks the sign of a single number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP ABS SWAP 0<
;

: D-SIGN	\ d -- |d| flag
\ *G Checks the sign of a double number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP 0< DUP >R
  IF  DNEGATE  THEN
  R>
;

: T-SIGN	\ t -- |t| flag
\ *G Checks the sign of a triple number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP 0<  dup >R
  IF  TNEGATE  THEN
  R>
;

: Q-SIGN	\ q -- |q| flag
\ *G Checks the sign of a quadruple number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP 0<  dup >R
  IF  QNEGATE  THEN
  R>
;

: t0=		\ t -- flag
\ *G Return true if t=0.
  or or 0=
;

: t0<>		\ t -- flag
\ *G Return true if t<>0.
  or or 0<>
;

: q0=		\ q -- flag
\ *G Return true if q=0.
  or or or 0=
;

: q0<>		\ q -- flag
\ *G Return true if q<>0.
  or or or 0<>
;


\ ******************
\ *S Multiplications
\ ******************

: umds*		\ ud u -- ut
\ *G Unsigned mixed multiply.
\ ** Multiply double by single producing a triple.
\ long mul :  x1 x0
\              * y0
\         -----------
\  x0*y0 =    z1 z0
\  x1*y0 = z2 z1
\         -----------
\          z2 z1 z0
\         ===========
\         ===========
  ROT OVER UM*                          \ x0*y0=z1.z0
  2SWAP UM*                             \ x1*y0=z2.z1
  -ROT 0 ROT 0 D+                       \ z1+z1 = z1.z2'
  ROT +
;

: mds*		\ d n -- t
\ *G Signed multiply of a double by a single to
\ ** return a triple result.
  dup if
    2dup xor >r			\ sign of result
    abs -rot  dabs rot umds*
    r> 0<
    if  tnegate  endif
  else
    drop 2drop  0 0 0		\ multiplier 0, return zeroes
  then
;

: umdd*		\ ud1 ud2 -- uq
\ *G Unsigned multiply, two unsigned doubles to produce a quad
\ long mul :       d1.x1 d1.x0
\                * d2.x1 d2.x0
\          -------------------
\  1               q1.x1 q1.x0 = d1.x0 * d2.x0
\  2         q1.x2 q1.x1       = d1.x1 * d2.x0
\  3         q1.x2 q1.x1       = d1.x0 * d2.x1
\  4   q1.x3 q1.x2             = d1.x1 * d2.x1
\     ------------------------
\      q1.x3 q1.x2 q1.x1 q1.x0
\     ========================
  { | ud1[ 2 cells ] ud2[ 2 cells ] tmp[ cell ] -- uq }
  ud2[ 2!  ud1[ 2!
  ud1[ cell+ @  ud2[ cell + @ um* tmp[ !		\ ud1.lo * ud2.lo
  ud1[ @ ud2[ cell + @ um* 0				\ ud1.hi * ud2.lo
  tmp[ @ 0 0 t+
  ud1[ cell + @ ud2[ @ um* 0 t+				\ ud1.lo * ud2.hi
  0  ud1[ @ ud2[ @ um* t+ 				\ ud1.hi * ud2.hi
;

: umts*		\ ut u -- uq
\ *G Unsigned mixed multiply of triple by single to produce a quad.
\ long mul :  x2 x1 x0
\                 * y0
\          -----------
\ x0*y0 =        z1 z0
\ x1*y0 =     z2 z1
\ x2*y0 =  z3 z2
\          -----------
\          z3 z2 z1 z0
\          ===========
  swap >r  dup >r		\ -- x0 x1 y0 ; R: -- x2 y0
  rot um*			\ -- x1 (z1.z0)l (z1.z0)h ; R: -- x2 y0
  rot 0 swap r@ um*		\ -- (z1.z0)l (z1.z0)l 0 (z2.z1)l (z2.z1)h ; R: -- x2 y0
  d+				\ -- (z1.z0)l (z2.z1')l (z2.z1')h ; R: -- x2 y0
  0 r> r> um* d+		\ -- uq
;

: mts*		\ t n -- q
\ *G Signed mixed multiply of triple by single to produce a quad.
  dup if                                \ check multiplier for 0=
    2dup xor >r				\ sign of result
    abs >r tabs r> umts*
    r> 0<
    if  qnegate  endif
  else
    drop 3drop 0 0 0 0                  \ multiplier 0= return zeroes
  then
;


\ ************
\ *S Divisions
\ ************

synonym uds/mod	mu/mod \ ud1 u2 -- urem udquot
\ *G Unsigned divide of double dividend (64 bit) by a single
\ ** divisor (32 bit), returning a single remainder and a
\ ** double quotient.
synonym ud/mod mu/mod	\ ud1 u2 -- urem udquot
\ *G A synonym for *\fo{mu/mod) that appears in some other Forth
\ ** systems.

: mds/rem	\  d n -- rem dquot
\ *S Signed 64 bit divided by 32 bit divisor, returning a single
\ ** remainder and a double quotient. No protection is
\ ** provided. Symmetric division. A signed version of *\fo{MU/MOD}.
  over >r				\ sign of remainder=sgn(dividend)
  2dup xor >r				\ sign of quotient
  abs -rot dabs rot mu/mod		\ unsigned divide
  r> 0<
  if  dnegate  endif
  r> 0<
  if  rot negate -rot  endif
;

: umts/mod	\ ut u -- urem utquot
\ *G Divide unsigned triple by unsigned single, returning unsigned
\ ** single remainder and quotient.
  >R                                    \ -- t0 t1 t2 ; R: -- u
  0 R@ UM/MOD                           \ -- t0 t1 r(t2/u) q(t2/u) ; R: -- u
  R> SWAP >R                            \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u)
  DUP >R                                \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u) u
  UM/MOD                                \  rem2.x1 = rem1 & quot1
  R> SWAP >R                            \ get divisor & save quot1 on return stack
  UM/MOD                                \ rem1.x0 = rem0 & quot0
  R> R>                                 \ retrieve quot 1 & quot2 from return stack
;

: umts/		\ ut u -- utquot
\ *G Divide unsigned triple by unsigned single, returning unsigned
\ ** single quotient.
  >R                                    \ -- t0 t1 t2 ; R: -- u
  0 R@ UM/MOD                           \ -- t0 t1 r(t2/u) q(t2/u) ; R: -- u
  R> SWAP >R                            \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u)
  DUP >R                                \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u) u
  UM/MOD                                \  rem2.x1 = rem1 & quot1
  R> SWAP >R                            \ get divisor & save quot1 on return stack
  UM/MOD nip                            \ rem1.x0 = rem0 & quot0
  R> R>                                 \ retrieve quot 1 & quot2 from return stack
;

: mts/rem	\ t n -- rem tquot
\ *S Signed divide a triple by a single, returning a single
\ ** remainder and a triple quotient. No protection is
\ ** provided. Symmetric division.
  over >r				\ sign of remainder=sgn(dividend)
  2dup xor >r				\ sign of quotient
  abs >r tabs r> umts/mod		\ unsigned divide
  r> 0<
  if  tnegate  endif
  r> 0<
  if  >r >r >r  negate  r> r> r>  endif
;

: mts/		\ t n -- tquot
\ *S Signed divide a triple by a single, returning a triple
\ ** quotient. No protection is provided. Symmetric division.
  2dup xor >r				\ sign of quotient
  abs >r tabs r> umts/			\ unsigned divide
  r> 0<
  if  tnegate  endif
;

code dt2*	\ t1 t2 -- t1' t2'
\ *G Shift left two triples by one bit. This is a factor used below.
\ ** The carry out from *\i{t1} is shifted into *\i{t2}.
  shl   #16 [ebp], # 1
  rcl   #12 [ebp], # 1
  rcl   #8 [ebp], # 1
  rcl   #4 [ebp], # 1
  rcl   #0 [ebp], # 1
  rcl   ebx, # 1
  next,
end-code

: utt/mod	\ ut1 ut2 -- utrem utquot
\ *G Unsigned divide of two triple numbers producing a triple
\ ** remainder and a triple quotient.
  {: | tdiv[ 3 cells ] trem[ 3 cells ] tquot[ 3 cells ] tmp -- :}
  tdiv[ T!  tquot[ T!   0 0 0 trem[ T!
  96 0 do
    trem[ 2 cells + @ 0< -> tmp		\ preserve carry out
    tquot[ T@ trem[ T@ dt2* trem[ T! tquot[ T!	\ double triple left shift
    trem[ T@ tdiv[ T@ tu< invert  	\ check trial subtract
    tmp or if
      trem[ T@ tdiv[ T@ t- trem[ T!	\ subtract
      1 tquot[ or!			\ update quotient l.s. bit
    endif
  loop
  trem[ T@  tquot[ T@
;

((
: D/MOD  D>F D>F FSWAP F2DUP F/ F>D 2DUP D>F F* F- F>D 2SWAP ;  ok
))


\ **********
\ *S CCSisms
\ **********

CCSisms? [if]		\ nz for CCS specifics
\ *G Change to zero not to compile CCS specific code.

: DPL-VALID?                                                    ( dpl --- bool )
   { %dpl --- }
   %dpl #128 < %dpl -128 > AND
  ;

: (dd+)		\ d1 d2 -- t
\ *G Add two signed doubles and return a signed triple.
  2 pick 0< -rot  dup 0<  t+
;

: (TT+)		\ t1 t2 --- q
\ *G Add two signed triad vakues and return a signed quad value.
  >r >r >r  dup 0<		\ t1 -> q1
  r> r> r>  dup 0<		\ t2 -> q2
  Q+
;

: (dd*)		\ d1 d2 -- t
\ *G Signed multiply of two doubles to produce a triple.
  2dup d0<> if
    d-sign >r
    2swap d-sign >r
    umdd* drop
    r> r> xor
    if  tnegate  then
  else
    2drop 2drop  0 0 0
  endif
;

: L.DPL>D.DPL	\ l.dpl -- d.dpl
\ *G Convert L.DPL to D.DPL
  OVER 0< SWAP
;

CREATE 10(*)	\ -- addr
\ *G Array of powers of ten.
\ 98765432109876543210
                     1 ,	\ 10^0
                    10 ,	\ 10^1
                   100 ,	\ 10^2
                  1000 ,	\ 10^3
                 10000 ,	\ 10^4
                100000 ,	\ 10^5
               1000000 ,	\ 10^6
              10000000 ,	\ 10^7
             100000000 ,	\ 10^8
            1000000000 ,	\ 10^9

: 10^N		\ N -- 10^N
\ *G Return value 10 to the power n, where n is in the range 0..9.
  N-SIGN >R   DUP #0 #9 WITHIN?
  IF  cells 10(*) + @  ELSE  DROP #0  THEN
  R> ?NEGATE				\ retrieve sign flag & adjust value
;

: N.10^		\ N -- 10exponent
\ *G Calculate positive 10exponent greater than N
  DUP IF
    abs 0				\ init exponent
    BEGIN
      1+ DUP 10^N
      2 PICK >				\ 10th exponent >= value
      OVER 9 >= OR			\ protect against overflow
    UNTIL
    NIP
  THEN
;

: D?DUP			\ D --- D\-1 or 0
  2DUP D0<> dup 0=
  if  nip nip  endif
;

: dn/mod	\  d n -- rem dquot
\ *S 64 bit divided by 32 bit divisor, returning a single
\ ** remainder and a double quotient.
\ ** N.B. If n is zero, rem and dquot are returned as zero.
  ?dup 0= if
    2drop 0 0 0
  else
    mds/rem  rot abs -rot
  endif
;

: DN/		\ d n -- d'
\ *G 64 bit signed value divide by 32 bit signed value.
\ ** If divisor=0 return 0 result.
\ ** N.B. The value will be rounded to -infinity or +infinity.
  DUP IF				\ check divisor for 0=
    OVER 0< OVER 0< XOR >R		\ calc and save result sign flag
    DUP >R                              \ save divisor
    DN/MOD                              \ calc rem0 quot0 quot1
    ROT 1+ R> ABS 2/ > IF               \ rem0 greater than 1/2 divisor
      DABS                              \ absolute value
      1 0 D+                            \ increment value
      R@ ?DNEGATE                       \ correct the sign
    THEN
    r> drop                             \ drop saved sign flag
  ELSE \ yes!
    DROP 2DROP 0 0
  THEN
;

Synonym D?	D0<>	\ d -- flag
Synonym D=D	D=	\ d1 d2 -- flag
Synonym D<D	D<	\ d1 d2 -- flag
Synonym D-ABS	dabs    \ d -- |d|

Synonym udn*	umds*	\ ud u -- ut
\ *G Unsigned mixed multiply. Multiply double by single to produce
\ ** a triple.
synonym DN*	mds*	\ d n -- t
\ *G Multiply double signed value by single signed value and return
\ ** triple signed result.
Synonym utn*	umts*	\ ut u -- uq
\ *G Unsigned mixed multiply. Multiply triple by single to produce
\ ** a quad.
Synonym tn*	mts*	\ t n -- q
\ *G Signed mixed multiply. Multiply triple by single to produce
\ ** a quad.
Synonym UTN/MOD umts/mod	\ ut n -- urem utquot
\ *G Unsigned divide of triple by single.

: TN/MOD	( t n -- rem tquot )
\ *G Protected divide: signed triple by single.
\ ** Check for zero divisor and return 0 answer.
  DUP 0=   				\ divide by 0 check
  IF 4DROP #0 #0 #0 #0  exit  THEN
  N-SIGN >R				\ ? sign & abs divisor save flag
  >R T-SIGN R> SWAP >R                  \ ? sign of BD & save flag
  DUP 1 = IF                            \ check divisor for 1=
    DROP  >R 0 -ROT R>                  \ put remainder of zero on stack
  ELSE
    umts/mod
  THEN
  R> R> XOR                             \ retrieve sign flags & calc new sign
  IF  TNEGATE  THEN                     \ adjust BD to sign flag
;

: TN/	\ t n -- tquot
\ *G Protected divide: signed triple by single.
\ ** Check for zero divisor and return 0 answer.
\ ** NOTE: The value will be rounded to -infinity or +infinity
  DUP 0=                   	\ divide by 0 check
  IF 4DROP #0 #0 #0 exit THEN
  N-SIGN >R				\ ? sign & abs divisor save flag
  >R T-SIGN R> SWAP >R                  \ ? sign of BD & save flag
  DUP 1 = IF                            \ check divisor for 1=
    DROP
  ELSE
    dup >r				\ preserve divisor
    umts/mod				\ -- urem utquot ; R: -- s1 s2 divisor
    >r rot r> swap			\ -- utquot urem ; R: -- s1 s2 divisor
    1+  r> 1 rshift u>			\ needs rounding?
    if  1 0 0 t+  endif
  THEN
  R> R> XOR                             \ retrieve sign flags & calc new sign
  IF  TNEGATE  THEN                     \ adjust BD to sign flag
;

: udd/mod	\ ud1 ud2 -- udrem udquot
\ *G Unsigned divide of two double numbers producing a double
\ ** remainder and a double quotient.
  {: | ddiv[ 2 cells ] drem[ 2 cells ] dquot[ 2 cells ] tmp -- :}
  ddiv[ D!  dquot[ D!  0 0 drem[ D!
  64 0 do
    drem[ 1 cells + @ 0< -> tmp		\ preserve carry out
    dquot[ D@ drem[ D@ q2* drem[ D! dquot[ D!	\ double double left shift
    drem[ D@ ddiv[ D@ du< invert  	\ check trial subtract
    tmp or if
      drem[ D@ ddiv[ D@ d- drem[ D!	\ subtract
      1 dquot[ or!			\ update quotient l.s. bit
    endif
  loop
  drem[ D@  dquot[ D@
;

: UTD/MOD	\ ut1 ud1 --- udrem utquot
\ *G Unsigned divide of a triple by a double.
 { | %t1[ 3 CELLS ] %drem[ 2 CELLS ] %tq[ 3 CELLS ] %d1[ 2 CELLS ] }
   %d1[ D!  %t1[ T!
   %t1[ T@ ROT DROP			\ t1.x2 t1.x1
   %d1[ D@ udd/mod #0 -ROT %tq[ T!
   DROP
   %t1[ T@ 2DROP SWAP %d1[ D@ udd/mod 2SWAP %drem[ D!
   #0 %tq[ T@ T+ %tq[ T!
   %drem[ D@ %tq[ T@
  ;

: TD/		\ t1 d2 -- t2
\ *G Protected divide: signed triple by double producing a triple.
\ ** Check for zero divisor and return 0 answer.
\ ** NOTE: The value will be rounded to -infinity or +infinity
  { | %d1[ 2 cells ] %t1[ 3 cells ] st1 sd2 -- }
  2dup d0<> if
    dup 0=				\ check for division by +ve single
    if  drop TN/  exit  endif
    d-sign -> sd2  %d1[ D!
    t-sign -> st1  %t1[ T!
    %t1[ T@ %d1[ D@ UTD/MOD %t1[ T!
    1 0 d+  %d1[ D@ d2/  2swap du< if	\ check for rounding
      %t1[ T@ 1 0 0 t+
    else
      %t1[ T@
    endif
    st1 sd2 xor
    if  tnegate  endif
  else
    2drop 3drop  0 0 0
  endif
;

: TD>RND                                                       ( t1 --- d1 dpl )
   \ *G Down round t1 value to d1 value and return dpl adjustment
   T-SIGN >R
   #0                          \ prime dpl adjustment t1.x0 t1.x1 t1.x2 dpl+ ...
   BEGIN
        OVER N.10^       \ calculate 10 exponent t1.x0 t1.x1 t1.x2 dpl+ dpl2 ...
        TUCK      \ save current adjustment t1.x0 t1.x1 t1.x2 dpl2 dpl+ dpl2 ...
        +           \ add to previous adjustment t1.x0 t1.x1 t1.x2 dpl2 dpl+ ...
        >R                    \ save total adjustment t1.x0 t1.x1 t1.x2 dpl2 ...
        DUP 0<>                                     \ any adjustment required ??
        IF   10^N TN/ False                   \ yes! adjust and go another round
        ELSE DROP True
        THEN
        R>
        SWAP
   UNTIL
   NIP                                                    \ drop zero t.x2 value
   OVER 0<                                                     \ d.x1 msb set ??
   IF 1+ >R #0 #10 TN/ DROP R> THEN             \ inc adjustment and shift value
   R>                                                 \ sign of the value set ??
   IF -ROT DNEGATE ROT THEN
;

: UQN/MOD	\ uq u -- urem uqquot
\ *G Protected divide: signed 128 bit quad by 32 bit single producing a single
\ ** remainder and a quad quotient.
\ ** Check for zero divisor and return 0 answer.
   {: | %divisor -- :}
   \ ------------------------
   dup to %divisor IF
        %divisor 1 =                                        \ check divisor 1= ??
        IF
             >r >r >r >r  0  r> r> r> r>                 \ remainder = zero
        ELSE
             0 %divisor UM/MOD                             \ 0.x3 = rem2 & quot3
             >R                                     \ save quot3 to return stack
             %divisor UM/MOD                           \  rem2.x1 = rem1 & quot1
             >R                                     \ save quot1 on return stack
             %divisor UM/MOD                            \ rem1.x0 = rem0 & quot0
             R> R>                   \ retrieve quot 1 & quot2 from return stack
        THEN
   ELSE
        drop 4DROP 0 0 0 0 0
   THEN ;

: QN/MOD	\ q n -- rem qquot
\ *G Protected divide: signed quad by single producing a single
\ ** remainder and a quad quotient.
\ ** Check for zero divisor and return 0 answer.
   {: | %divisor %sflag -- :}
   \ ------------------------
   dup to %divisor IF
        %divisor N-SIGN to %sflag to %divisor         \ ? sign & abs divisor
        Q-SIGN >R                                      \ ? sign of Q & save flag
        %divisor 1 =                                        \ check divisor 1= ??
        IF
             >r >r >r >r  0  r> r> r> r>                 \ remainder = zero
        ELSE
             0 %divisor UM/MOD                             \ 0.x3 = rem2 & quot3
             >R                                     \ save quot3 to return stack
             %divisor UM/MOD                           \  rem2.x1 = rem1 & quot1
             >R                                     \ save quot1 on return stack
             %divisor UM/MOD                            \ rem1.x0 = rem0 & quot0
             R> R>                   \ retrieve quot 1 & quot2 from return stack
        THEN
        R> R> XOR                          \ retrieve sign flags & calc new sign
        IF QNEGATE THEN                                 \ adjust BD to sign flag
   ELSE
        drop 4DROP 0 0 0 0 0
   THEN ;

: QN/		\ q n -- qquot
\ *G 128bit signed value divide with 32bit signed value.
\ ** check for 0= divisor and return 0 answer
  QN/MOD 4 roll drop
;

Synonym TVAR@ T@
Synonym TVAR! T!

' T@ >name hidename
' T! >name hidename

: D.DPL-NEGATE                                                     ( b1 --- b2 )
\ *G Negate a double signed number with a dpl
   >R DNEGATE R> ;

: L.DPL-NEGATE                                              ( N\DPL --- -N\DPL )
\ *G Negate a single signed number with a dpl
   SWAP NEGATE SWAP ;

: L.DPL-SIGN                                                ( S --- S abs\flag )
\ *G Checks the sign of a single number with dpl and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
   >R N-SIGN R> SWAP ;

: D.DPL-SIGN                                        ( D.DPL --- ABS_D.DPL\flag )
\ *G Checks the sign of a double number with dpl and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
   >R D-SIGN R> SWAP
  ;

: D.DPL-ABS                                              ( D.DPL --- abs D.DPL )
\ *G Return the absolute of a double number with dpl
   >R DABS R> ;

: D.DPL>T.DPL                                                ( d.dpl --- t.dpl )
\ *G Convert a D.DPL to a T.DPL
   OVER 0< SWAP ;

: D.DPL>$                                                 ( d.dpl --- addr cnt )
   \ *G Convert d.dpl to decimal string
   BASE @ >R                                                 \ save current base
   DECIMAL                                              \ set base 10 conversion
   D.DPL-SIGN >R                                                \ save sign flag
   <#                                         \ start picture numeric conversion
   DUP 0>
   IF   #0 DO # LOOP
        [CHAR] . HOLD
   ELSE ABS #0 #0 ROT #0 ?DO # LOOP
        2DROP
   THEN
   #S                                    \ process balance of numeric conversion
   R> SIGN #>                                        \ get sign flag and process
   R> BASE ! ;                                            \ restore current base

: T#   ( ut --- ut' )
\ *G Given a tripple number on the stack this will add the next digit to
\ ** the pictured numeric output buffer and return the next tripple
\ ** number to work with. N.B. the output string is built from
\ ** right (lsd) to left (msd).
   BASE @ UMTS/MOD 2SWAP SWAP
   9 OVER < IF  7 +  THEN
   [CHAR] 0 + HOLD
   -ROT
;

: T>$                                                  ( x0 x1 x3 --- addr cnt )
   T-SIGN >R
    <#                                        \ start picture numeric conversion
   BEGIN
       #10 UMTS/MOD 2SWAP SWAP
       [CHAR] 0 + HOLD
       -ROT
       3DUP OR OR 0<>                                            \ zero 96bit ??
       IF   False
       ELSE True
       THEN
   UNTIL
   3DROP
   R> SIGN                                           \ get sign flag and process
   HLD @ PAD OVER -
  ;

[then]


\ *********************
\ +S Protected division
\ *********************
\ +P Protected division can be used for two reasons
\ +( 1
\ +B Avoiding problems caused by unused data cells
\ +* being set to zero.
\ +B Avoiding overflow exceptions
\ +)

\ +P In many applications, the first  case can be handled by
\ +* testing for a zero divisor and returning zero for quotient
\ +* and/or remainder. The second case is almost always an error.

((
: */modP  \ n1 n2 n3 -- rem quot ; protected version
  -rot m* rot                   \ -- d n3
  over abs
  over abs 1 rshift  >= if      \ |dh| >= |n3| on error
    drop 2drop 0 0
  else
    sm/rem
  endif
;
))

\ ======
\ *> ###
\ ======

decimal

