\ VFX Forth hardware floating point support

((
Copyright:
----------
Copyright (c) 1986, 1987, 1998, 1999, 2000, 2002, 2004, 2007, 2008 to:
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England
tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: tech-support@mpeforth.com
web: http://www.mpeforth.com

N.B. Bugs in the assembler were fixed for v4.x. Do not use this file
     with versions before v4.0.
N.B. The bug in the FSUB/P and FSUBR/P assembler instructions has
     been fixed as of build 0785. Use them with care!

To do
=====
Remove all uses of PAD.

According to Forth-94:

    12.6.1.2143 REPRESENT     FLOATING

    ( c-addr u -- n flag1 flag2 )  (F: r -- )   or
      ( r c-addr u -- n flag1 flag2 )

    At c-addr, place the character-string external representation
    of the significand of the floating-point number r.  ...
    The character string shall consist of the u most significant
    digits ...  The significand is rounded to u digits following the
    round to nearest rule ...

The above states the number must be *rounded* to u digits, so

    1234.56E0  PAD 5 REPRESENT
    PAD 5 DUMP
    0CDF:2352  31 32 33 34 36   12346 ok

giving the resulting string  '12346'.

Unfortunately some forths have  REPRESENT  truncate and so
give  '12345'.  This is non-compliant and will cause incorrect
number display in applications that use  REPRESENT.


Change history
==============
20090429 SFP027 Updated for tokeniser.
20080923 SFP026 Corrected trig functions.
20080922 SFP025 Corrected rounding errors when generating exponent.
20070103 EDA024 Updated REPRESENT
20040701 SFP023	Restored checked version of F** as the FSL requires it.
20021023 SFP022 Added FR>S and FR>D after much discussion on comp.lang.forth
20021007 SFP021 Corrected F>S and F>D.
SFP020	20 Mar 2002	Corrected F>ASCII and >FLOAT
SFP019	18 Feb 2002	Corrected REPRESENT and rounding control in >BCD.
			Corrected exponent handling in numeric output.
SFP018                  dummy
SFP017	13 Mar 2001	Added F# for PFW2.x compatibility
SFP016	2 Jan 2000	Improved rounding during printing in >BCD.
			Renamed (FLITERAL) to (RLITERAL) to avoid name conflict.
			Added REALS and INTEGERS for user control of number input.
SFP015	25 Sept 2000	Changed >BCD to preserve temp value as 10 byte real.
			This stops 1E400 being displayed as 9.9999e399 etc.
SFP014	23 Sept 2000	Corrected 2**z to preserve EDX
SFP013	20 Sept 2000	Added FPSYSTEM value to define FP type in use
SFP012	19 Sept 2000	Corrected display for NaNs and Infs
SFP011	24 July 2000	Corrected FMOD
SFP010	20 July 2000	Inlined some code definitions.
			Renamed popFPU and pushFPU to FPU> and >FPU
SFP009	10 Oct 1999	Corrected rounding control in I/O operations.
			Made 2**Z general, Z may be negative.
			Clipped F** ( F1^F2 ) for F1<=0.
			Enhanced prompt for floats.
			Added FP debugger.
SFP008	9 Oct 1999	Added FVALUE
SFP007	8 Oct 1999	Fixed F~
SFP006  7 Oct 1999      Renamed FULL-NUMBER? to FNUMBER?
			Changed FULL-NUMBER? to check for integers first
			Enhanced the documentation section.
			Changed number output to use FP-CHAR for point separator
SFP005  28 Sept 1999	FST cannot use 80 bit real to memory!
SFP004  15 Sept 1999	Tidy ups and code improvements (unmarked)
SJC005  99/03/08 	Modified FNUMBER?, can only parse float number if BASE is
                	DECIMAL according to ANS
SJC004  99/03/02 	Fixed FPICK after SFP002. Please note that *4 is <<2 not <<3
                	and the *8 is <<3 not <<2 - snigger!
SFP003  11 Feb 1999     Modified for VFX Forth
SFP002  5 Aug 1998      Improved FOVER, FPICK and miscellaneous words
SFP001  1 Aug 1998      Modified for X8632 cross compiler
))

( *! float     )
( *T Floating Point )

( *S Introduction )
( *P The packages use the FPU instruction set. The source code can be found )
( ** in the files LIB\HFP387.FTH, LIB\NDP387.FTH and LIB\HFPOGL32.FTH. )

( *N HFP387.FTH - external FP stack )
( *P In HFP387.FTH floating point numbers are kept on a separate stack, )
( ** pointed to by the USER variables FSP and FS0. The top of the FP stack is )
( ** cached in the FPU. )

( *P All tasks, winprocs and callbacks are allocated a separate 4096 byte )
( ** floating point stack. If you need a larger one, allocate it from the )
( ** heap using ALLOCATE, and modify FSP and FS0 accordingly. Note that the )
( ** stack grows down. )

( *P By default, HFP387.FTH defines floating point stack items and )
( ** literals to be in 80 bit extended real format. If you need to save )
( ** space, the default can be changed by setting the constant FPCELL to )
( ** 8 for 64 bit double precision or to 4 for 32 bit single precision. )
( ** If you do this, accuracy and resolution may/will suffer. )

( *N NDP387.FTH - coprocessor stack
( *P In NDP387.FTH floating point numbers are kept in the floating point )
( ** unit's internal stack only. This code is significantly faster, but )
( ** is limited to the use of 8 floats on the NDP stack, including working )
( ** temporary numbers. )

( *P NDP387.FTH defines literal and default floats using )
( ** F@ F! and friends to be in 80 bit format. Change the constant FPCELL )
( ** as above to use a different default size. )

( *N HFPOGL32.FTH - 32 bit floats on data stack )
( *P This file is specifically designed for interfacing to OpenGL APIs. )
( ** Floating point numbers are kept as 32 bit floats on the data stack. )
( ** Compared to the previous two packages, HfpOGL32 has a restricted range )
( ** of facilities and is a subset of the other two packages. See the source )
( ** code for more details. )

\ Note that the multitasker and interrupt routines must save and restore )
\ these variables through context switches.

( *S Radians and Degrees )
( *P Please note that the trig functions are calculated in )
( ** radians, for calculations in degrees use DEG>RAD beforehand, )
( ** RAD>DEG afterwards. )

( *S Number formats and ANS Forth )
( *P The ANS Forth standard specifies that floating point numbers must )
( ** be entered in the form 1.234e5 and must contain a point '.' and )
( ** 'e' or 'E', and that double integers are terminated by a point '.'. )

( *P This situation prevents the use of the standard conversion words )
( ** in international applications because of the interchangable use of )
( ** the '.' and ',' characters in numbers. Because of this, VFX Forth )
( ** uses two system variables, FP-CHAR and DP-CHAR, to hold the character )
( ** used as the floating point and double integer indicator characters. )
( ** By default, FP-CHAR is initialised to '.' and DP-CHAR is initialised )
( ** to ',', as has been MPE practice for many years. For ANS compliance, )
( ** you should set them as follows: )

( *E \ ANS standard setting     )
( **   char . dp-char !		)
( **   char . fp-char !	        )
( ** : ans-floats	\ -- ; for strict ANS compliance        )
( **   [char] . dp-char !	)
( **   [char] . fp-char !	)
( ** ;				)
( ** \ MPE defaults	        )
( **   char , dp-char !		)
( **   char . fp-char !		)
( ** : mpe-floats	\ -- ; for existing and legacy code      )
( **   [char] , dp-char !	)
( **   [char] . fp-char !	)
( ** ;				)

( *P You can of course set these variables to any value which suits you )
( ** for your language and locale. Note that integer conversion is always )
( ** attempted before floating point conversion. This means that if FP-CHAR )
( ** and DP-CHAR are the same, floating point numbers must contain 'e' or )
( ** 'E'. If they are different, a number containg the FP-CHAR will be )
( ** successfully converted as a floating point number, even if it does )
( ** not contain 'e' or 'E'. )

( *S Floating point exceptions )
( *P Exception handling is determined by the operating system. On current )
( ** Windows platforms, floating point exceptions are not generated by )
( ** The NDP. This can be changed by altering the bottom 6 bits of the )
( ** NDP control word using CW@ and CW!. )

( *P By default, the system prompt will report exception status, and clear )
( ** the pending exception status. Exception status reporting does not mean that )
( ** the Windows exception handler has been triggered, it only means that )
( ** the status flag has been set. )

\ *S Standards compliance, F>S and F>D.
\ *P After much discussion on the comp.lang.forth newsgroup, a consensus
\ ** was reached that F>D and F>S must truncate to zero. This is also
\ ** the behaviour required by the Forth Scientific Library (FSL). Historically,
\ ** MPE floating point packs permit the integer rounding mode to be set
\ ** by the user. In order to support both camps, VFX Forth now behaves as
\ ** follows:
\ *B F>D and F>S truncate to zero,
\ *B FR>D and FR>S follow the current rounding mode.


( *S Glossary )


\ *******************************
\ Floating point type definitions
\ *******************************

[undefined] FpSystem [if]
0 value FpSystem
( *G The value FPSYSTEM defines which floating point pack is installed and )
( ** active. Each floating point pack defines its own type as follows: )
( *B 0 constant NoFPSystem )
( *B 1 constant HFP387System )
( *B 2 constant NDP387System )
( *B 3 constant OpenGL32System )
[then]

\ 0 constant NoFPSystem
  1 constant HFP387System
\ 2 constant NDP387System
\ 3 constant OpenGL32System
HFP387System -> FpSystem

\ Define size of items on the floating point stack, 4, 8, or 10 bytes
#10 constant FPCELL	\ -- n
( *G Defines the size of literals and floating point numbers in memory )
( ** and on the floating point stack. FPCELL can be changed to )
( ** 8 for 64 bit double precision or to 4 for 32 bit single precision. )
( ** If you do this, accuracy and resolution may/will suffer. )
FPCELL negate constant -FPCELL

\ Define number of output digits
FPCELL #10 = [if] #18 [else]
FPCELL 8   = [if] #16 [else]
FPCELL 4   = [if] 7   [else]
  0  cr ." Invalid FPCELL value" abort
[then] [then] [then]
  constant #fdigits	\ -- u
\ *G Returns the largest number of usable digits available from
\ ** *\fo{REPRESENT}. Equivalent to then environment variable
\ ** *\fo{MAX-FLOAT-DIGITS}.


\ ****************************
\ Floating point debug support
\ ****************************

false constant [fpdebug] immediate
( *G Set this CONSTANT true when compiling HFP387.FTH, and a debug )
( ** build will be constructed. In this state, the state of the )
( ** FPU is checked after each word. If a floating point exception )
( ** has been generated, a diagnostic is issued, and the system )
( ** aborts. Set this only when testing. )

[fpdebug] [if]
  defer fpcheck				\ see later for real action
  assign noop to-do fpcheck
[then]


\ *********************************
\ Variable and frames for FP stacks
\ *********************************

\ memory copies of the staus word, control word, and rounding control word
\ are held on a stack frame built as an offset on the return stack pointed
\ to by EBP.
0 constant #sw          \ offset of status word
2 constant #cw          \ offset of control word
4 constant #rc          \ offset of rounding control word
6 constant #spare       \ dummy to bring up to 8 bytes

8 constant #fpframe     \ size of frame required for all four items


\ ****************
\ Assembler macros
\ ****************

also asm-access definitions

fpcell 4 = [if]
: fword		\ --
( *G Selects appropriate floating point size for the assembler. )
( ** Note that this is defined by the constant FPCELL. )
( ** FWORD will be a synonym for FLOAT, DOUBLE or TBYTE. )
  float  ;
[then]
fpcell 8 = [if]
: fword		\ --
  double  ;
[then]
fpcell #10 = [if]
: fword		\ --
  tbyte  ;
[then]

: >FPU       \ -- ; push NOS on FP stack to FPU
  mov   eax, FSP-OFFSET [esi]   \ get FP stack pointer
  fld   fword 0 [eax]           \ push FNOS to FPU
  lea   eax, FPCELL [eax]	\ update stack pointer
  mov   FSP-OFFSET [esi], eax
;

: FPU>        \ -- ; pop FTOS on FPU to FP stack
  mov   eax, FSP-OFFSET [esi]   \ get FP stack pointer
  lea	eax, -FPCELL [eax]	\ update stack pointer
  fstp  fword 0 [eax]           \ store and pop
  mov   FSP-OFFSET [esi], eax
;

: fnext,	\ -- ; can be changed for debugging
[fpdebug] [if]
  fwait				\ forces last instruction flush
  call 	[] ['] fpcheck >body
[then]
  ret
;

previous definitions


\ ************
\ FP constants
\ ************

code %0         \ F: -- f#(0)
( *G Floating point 0.0 )
  FPU>
  fldz
  fnext,
end-code

code %1         \ F: -- f#(1)
( *G Floating point 1.0 )
  FPU>
  fld1
  fnext,
end-code

code %pi        \ F: -- f#(pi)
( *G Floating point PI )
  FPU>
  fldpi
  fnext,
end-code

code %pi/2      \ F: -- f#(pi/2)
( *G Floating point PI/2 )
  FPU>
  fld1
  fadd  st, st
  fldpi
  fdivrp st(1), st
  fnext,
end-code

code %pi/4      \ F: -- f#(pi/4)
( *G Floating point PI/4 )
  FPU>
  fld1                          \ 1.0
  fadd  st, st                  \ 2.0
  fadd  st, st                  \ 4.0
  fldpi
  fdivrp st(1), st
  fnext,
end-code

code %lg2e	\ F: -- log2(e)
\ *G Returns log (base 2) of e.
  FPU>  fldl2e  fnext,  end-code


\ *********************
\ FP control operations
\ *********************

code (finit)	\ F: -- ; resets FPU
  finit
  fldz
  next,
end-code  inline

: finit		\ F: i*f -- ; resets FPU and FP stack
( *G Reset the floating point unit and the floating point stack. )
  (finit)
  fs0 @ fsp !
;

finit

code cw@	\ -- cw ; get NDP control word
( *G Return the floating point unit Control Word. )
  lea	ebp, -8 [ebp]		\ save TOS, reserve one more cell
  mov	4 [ebp], ebx
  fstcw 0 [ebp]			\ get control word to memory
  xor	ebx, ebx
  mov	bx, 00 [ebp]		\ and then to TOS
  lea	ebp, 4 [ebp]		\ release spare cell
  fnext,
end-code

code cw!	\ cw -- ; set NDP control word
( *G Set the floating point unit Control Word. )
  lea   ebp, -4 [ebp]		\ must load CW to memory
  mov	0 [ebp], ebx		\ new CW is in TOS
  fldcw 0 [ebp]
  mov	ebx, 4 [ebp]		\ restore TOS
  lea	ebp, 8 [ebp]
  fnext,
end-code

code sw@	\ -- sw ; get NDP status word
( *G Return the floating point unit Status Word. )
  lea	ebp, -4 [ebp]		\ save TOS, reserve one more cell
  mov	0 [ebp], ebx
  fstsw ax			\ get status word
  sub 	ebx, ebx		\ clear result
  mov	bx, ax
  next,				\ so that debugger isn't triggered repeatedly
end-code

code fclex	\ -- ; clear exceptions
( *G Clear any pending floating point exceptions. )
  fclex				\ clear exceptions
  next,
end-code


\ *******************
\ FP Stack operations
\ *******************

code fdup       \ F: f -- f f
( *G Floating point equivalent of DUP. )
  mov   eax, FSP-OFFSET [esi]
  sub   eax, # FPCELL
  mov   FSP-OFFSET [esi] , eax
fpcell #10 <> [if]		\ SFP005
  fst   fword 0 [eax]
[else]
  fld	st			\ duplicate TOS
  fstp  fword 0 [eax]		\ FST cannot store extended 80 bit
[endif]
  fnext,
end-code

code fswap      \ F: f1 f2 -- f2 f1
( *G Floating point equivalent of SWAP. )
  mov   eax, FSP-OFFSET [esi]
  fld   fword 0 [eax]
  fxch
  fstp  fword 0 [eax]
  fnext,
end-code

code fdrop      \ F: f --
( *G Floating point equivalent of DROP. )
  fstp  st
  >FPU
  fnext,
end-code

code fover      \ F: f1 f2 -- f1 f2 f1
( *G Floating point equivalent of OVER. )
  FPU>
  fld   fword FPCELL [eax]
  fnext,
end-code

code frot       \ F: f1 f2 f3 -- f2 f3 f1
( *G Floating point equivalent of ROT. )
  mov   eax, FSP-OFFSET [esi]
  fld   fword 0 [eax]                   \ get f2, to be replaced by f3
  fxch
  fstp  fword 0 [eax]                   \ store f3 where f2 was
  fld   fword FPCELL [eax]              \ get f1, to be replaced by f2
  fxch
  fstp  fword FPCELL [eax]              \ store f2 where f1 was, f1 in FTOS
  fnext,
end-code

code fpick      \ n -- ; F: -- f
( *G Floating point equivalent of PICK. Note that because the )
( ** pick index is an integer, it is on the normal Forth integer )
( ** data stack, and the result, being a floating point number, )
( ** is on the floating point stack. )
  or    ebx, ebx                        \ check range
  s, if,  sub   ebx, ebx  endif,        \ force 0 if negative
  FPU>                                \ preserve old FTOS, leaves FP in EAX
fpcell 4 = [if]
  fld   fword 0 [eax] [ebx*4]		\ get f from deep stack
[then]
fpcell 8 = [if]
  fld   fword 0 [eax] [ebx*8]           \ get f from deep stack
[then]
fpcell #10 = [if]
  shl   ebx, # 1                        \ convert to index, by shift and add
  lea   ebx, 0 [ebx] [ebx*4]		\ *2 + *8 = *10
  fld   fword 0 [eax] [ebx]		\ get f from deep stack
[then]
  mov   ebx, 0 [ebp]                    \ restore TOS
  lea   ebp, 4 [ebp]
  fnext,
end-code

((
\ From Andrew Haley on comp.lang.Forth
\ This has a side effect on PAD, unfortunately.  However, it's portable
\ and multi-tasking safe.
: froll		\ n -- ; F: fn..f0 -- fn-1..f0 fn
  ?dup 0= if  exit  then
  >r  pad r@ floats + f!
  r@ 1- recurse
  pad r> floats + f@
  fswap
;
))

code ndepth     \ -- n ; depth of NDP stack
( *G Returns on the Forth data stack the number of items on the )
( ** FPUs's internal working stack. )
  fstsw ax                              \ NDP status
  sub   ebp, # 4                        \ preserve TOS
  mov   0 [ebp], ebx
  shr   ax, # #11                       \ line up the bits
  and   al, # $07                       \ only want bottom three
  mov   ebx, # 8                        \ 8 registers on NDP,
  sub   bl, al                          \ convert to depth
  cmp   bl, # 8                         \ convert 8 to 0
  z, if,  sub bl, bl  endif,
  fnext,
end-code

: fdepth        \ -- #f
( *G Floating point equivalent of DEPTH. The result is reurned )
( ** on the Forth data stack, NOT the float stack. )
  FS0 @  fsp @ -  FPCELL /              \ possible # of floats on FP stack
;


\ *************************************
\ Memory operations SF@ SF! DF@ DF! etc
\ *************************************

code f@         \ addr -- ; F: -- f
( *G Places the contents of addr on the float stack. The size )
( ** of the item fetched was defined by FPCELL. )
  FPU>                        \ preserve previous FTOS
  fld   fword 0 [ebx]           \ load from the address
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code sf@        \ addr -- ; F: -- f
( *G Places the 32 bit float at addr on the float stack. )
  FPU>                        \ preserve previous FTOS
  fld   float 0 [ebx]           \ load from the address
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code df@        \ addr -- ; F: -- f
( *G Places the 64 bit double float at addr on the float stack. )
  FPU>                        \ preserve previous FTOS
  fld   double 0 [ebx]          \ load from the address
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code tf@        \ addr -- ; F: -- f
( *G Places the 80 bit extended float at addr on the float stack. )
  FPU>                        \ preserve previous FTOS
  fld   tbyte 0 [ebx]		\ load from the address
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code f!         \ addr -- ; F: f --
( *G Stores the top of the float stack as an FPCELL sized )
( ** number at addr. )
  fstp  fword 0 [ebx]
  >FPU                       \ restore FTOS from deep stack
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code sf!        \ addr -- ; F: f --
( *G Stores the top of the float stack as an 32 bit float )
( ** number at addr. )
  fstp  float 0 [ebx]
  >FPU                       \ restore FTOS from deep stack
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code df!        \ addr -- ; F: f --
( *G Stores the top of the float stack as an 64 bit double float )
( ** number at addr. )
  fstp  double 0 [ebx]
  >FPU                       \ restore FTOS from deep stack
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code tf!        \ addr -- ; F: f --
( *G Stores the top of the float stack as an 80 bit extended float )
( ** number at addr. )
  fstp  tbyte 0 [ebx]
  >FPU                       \ restore FTOS from deep stack
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code inline

code f+!	\ F: f -- ; addr -- ; add f to data at addr
( *G Add F to the data at ADDR. )
  fld   fword 0 [ebx]		\ get from addr
  faddp st(1), st               \ do the actual addition
  fstp  fword 0 [ebx]		\ put back at addr
  >FPU                       \ restore FTOS from deep stack
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code

code f-!	\ F: f -- ; addr -- ; sub f from data at addr
( *G Subtract F from the data at ADDR. )
  fld   fword 0 [ebx]		\ get from addr
  fsubrp st(1), st		\ do the actual subtraction
  fstp  fword 0 [ebx]		\ put back at addr
  >FPU                       \ restore FTOS from deep stack
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code

\ ************************
\ *S Dictionary operations
\ ************************

: tf,           \ F: f --
( *G Lays an 80 bit extended float into the dictionary, reserving )
( ** 10 bytes )
  here  #10 allot  tf!          \ allocate 10 bytes and stash number here
;

: df,           \ F: f --
( *G Lays an 64 bit double float into the dictionary, reserving )
( ** 8 bytes )
  here  8 allot  df!            \ allocate 8 bytes and stash number here
;

: sf,           \ F: f --
( *G Lays a 32 bit float into the dictionary, reserving )
( ** 4 bytes )
  here  4 allot  sf!            \ allocate 4 bytes and stash number here
;

: f,            \ F: f --
( *G lays a default float into the dictionary, reserving )
( ** FPCELL bytes )
  here  FPCELL allot  f!  ;

: falign        \ --
( *G Aligns the dictionary to accept a default float. )
; immediate

: faligned      \ addr -- addr'
( *G Aligns the address to accept a default float. )
;

: float+        \ addr -- addr'
( *G Increments addr by FPCELL, the size of a default float. )
  FPCELL +  ;

: floats        \ n1 -- n2
( *G Returns n2, the size of n1 default floats. )
  FPCELL *
;

: sfalign       \ --
( *G Aligns the dictionary to accept a 32 bit float. )
; immediate

: sfaligned     \ addr -- addr'
( *G Aligns the address to accept a 32 bit float. )
; immediate

: sfloat+       \ addr -- addr'
( *G Increments addr by the size of a 32 bit float. )
  4 +
;

: sfloats       \ n1 -- n2
( *G Returns n2, the size of n1 32 bit floats. )
  4 *
;

: dfalign       \ --
( *G Aligns the dictionary to accept a 64 bit double float. )
; immediate

: dfaligned     \ addr -- addr'
( *G Aligns the address to accept a 64 bit float. )
; immediate

: dfloat+       \ addr -- addr'
( *G Increments addr by the size of a 64 bit double float. )
  8 +
;

: dfloats       \ n1 -- n2
( *G Returns n2, the size of n1 64 bit double floats. )
  8 *
;

: tfalign       \ --
( *G Aligns the dictionary to accept an 80 bit extended float. )
; immediate

: tfaligned     \ addr -- addr'
( *G Aligns the address to accept an 80 bit extended float. )
; immediate

: tfloat+       \ addr -- addr'
( *G Increments addr by the size of an 80 bit extended float. )
  #10 +
;

: tfloats       \ n1 -- n2
( *G Returns n2, the size of n1 80 bit extended floats. )
  #10 *
;


\ *****************
\ FP defining words
\ *****************

: fvariable     \ F: -- ; -- addr
( *G Use in the form: FVARIABLE <name> to create a variable )
( ** that will hold a default floating point number. )
  FPCELL buffer:
;

: farray                \ n -- ; i -- addr
( *G Use in the form: n FARRAY <name> to create a variable    )
( ** that will hold a default floating point number. When the )
( ** array name is executed, the index i is used to retun the )
( ** address of the i'th 0 zero-based element in the array.   )
( ** For example, 5 FARRAY TEST will set up 5 array elements  )
( ** each containing 0, and then f n TEST F! will store f in  )
( ** the nth element, and n TEST F@ will fetch it.            )
  create
    0 max  floats allot&erase
  does>
    swap FPCELL * +                     \ make address
;

: fconstant     \ F: f -- ; F: -- f
( *G Use in the form: <float> FCONSTANT <name> to create a constant )
( ** that will return a floating point number. )
  create
    f,
  does>
    f@
;

: fValComp,	\ xt --
\ Compiler for children of FVALUE.
  >body  [compile] literal
  case  OperatorType @off
    0 of  postpone f@  endof		\ Default, fetch
    1 of  postpone f!  endof		\ store, ->, to
    2 of  ( noop )  endof		\ addr
\					\ inc
\					\ dec
    5 of  postpone f+!  endof		\ add
\					\ zero
    7 of  postpone f-!  endof		\ sub
    8 of  postpone drop  postpone FPCELL  endof	\ sizeof
\ 					\ set
      bad-method
  endcase
; doNotSin

[-short-branches
: fvalue	\ F: f -- ; ??? -- ???
( *G Use in the form: *\fo{<float> FVALUE <name>} to create a floating )
( ** point version of VALUE that will return a floating point number )
( ** by default, and that can accept the operators TO, ADDR, )
( ** ADD, SUB, and SIZEOF. )
  create
    f,  ['] fValComp, set-compiler
  interp>
    case  OperatorType @off
      0 of  f@  endof			\ default, fetch
      1 of  f!  endof			\ store, ->, to
      2 of  ( noop )  endof		\ addr
\      3 of  incr  endof		\ inc
\      4 of  decr  endof		\ dec
      5 of  f+!  endof		\ add
\      6 of  off  endof		\ zero
      7 of  f-!  endof		\ sub
      8 of  drop FPCELL  endof	\ sizeof
\      9 of  on  endof		\ set
         bad-method
    endcase
;
short-branches]


\ *******************************************
\ FP four basic functions + - * /  and others
\ *******************************************

code f+         \ f1 f2 -- f1+f2
( *G Floating point add. )
  >FPU
  faddp st(1), st               \ do the actual addition
  fnext,
end-code inline

code f-         \ f1 f2 -- f1-f2
( *G Floating point subtract. )
  >FPU                       \ get f1 to FPU
  fsubrp st(1), st              \ do the actual calc.
  fnext,
end-code inline

code f*         \ f1 f2 -- f1*f2
( *G Floating point multiply. )
  >FPU                       \ get f1 to FPU
  fmulp st(1), st               \ do the actual calc.
  fnext,
end-code

code f/         \ f1 f2 -- f1/f2
( *G Floating point divide. )
  >FPU                       \ get f1 to FPU
  fdivrp st(1), st              \ do the actual calc.
  fnext,
end-code inline

code fmod       \ F: f1 f2 -- f3
( *G Floating point modulus. Returns f3 the remainder after repeatedly )
( ** subtracting f2 from f1. Often used to force arguments to lie in the )
( ** range: 0 <= arg < f2 )
  >FPU                       \ load f1
  fprem                         \ take modulus, rem in ST, f2 in ST(1)
  fxch
  fstp  st                      \ discard f2
  fnext,
end-code inline

code fsqrt      \ F: f -- sqrt(f)
( *G Floating point square root. )
  fsqrt
  fnext,
end-code inline

code 1/f        \ F: f -- 1/f
( *G Floating point reciprocal. )
  fld1                          \ push one on ap stack
  fdivrp st(1), st              \ do the actual division
  fnext,
end-code inline

code fabs       \ F: f -- |f|
( *G Floating point absolute. )
  fabs
  fnext,
end-code inline

code fnegate    \ F: f -- -f
( *G Floating point negate. )
  fchs
  fnext,
end-code inline

code f2*        \ F: f -- f*2
( *G Floating point multiply by two. )
  fld1
  fadd  st, st
  fmulp st(1), st
  fnext,
end-code inline

code f2/        \ F: f -- f/2
( *G Floating point divide by two. )
  fld1
  fadd  st, st
  fxch
  fdivrp st(1), st
  fnext,
end-code inline


\ ************************
\ Integer to FP conversion
\ ************************

code s>f        \ n -- ; F: -- f
( *G Converts a single integer to a float. )
  FPU>                \ put current FTOS on deep stack
  mov   eax, 0 [ebp]	\ swap TOS and NOS, so that TOS is in memory
  mov   0 [ebp], ebx
  mov   ebx, eax
  fild  dword 0 [ebp]   \ load 32 bit integer to FPU
  add   ebp, # 4        \ drop n from data stack
  fnext,
end-code

code d>f        \ d -- ; F: -- f
( *G Converts a double integer to a float. )
  FPU>                \ put current FTOS on deep stack
\ The double number needs to be in memory in strict Intel order
  mov   eax, 0 [ebp]    \ get low portion of double int, now in EBX:EAX
  sub   ebp, # 4        \ make space
  mov   0 [ebp], eax    \ stash double
  mov   4 [ebp], ebx
  fild  qword 0 [ebp]   \ load 64 bit integer to FPU
  mov   ebx, 8 [ebp]    \ restore TOS
  add   ebp, # #12      \ discard 64 bit integer and extras
  fnext,
end-code

\ fixed SFP021
code f>s        \ F: f -- ; -- n ; convert float to integer
\ *G Converts a float to a single integer.
\ ** Note that F>S truncates the number towards zero
\ ** according to the ANS specification.
\ ** See FR>S below.
  sub   ebp, # 8        	\ reserve space on stack
  fstcw word 0 [ebp]		\ save/preserve CW
  fstcw word 2 [ebp]
  or    word 0 [ebp], # $0C00	\ force truncation
  fldcw word 0 [ebp]
  fistp dword 4 [ebp]		\ store integer from FPU
  fldcw word 2 [ebp]
  mov   eax, 4 [ebp]		\ get new TOS
  mov   4 [ebp], ebx		\ save old TOS
  add   ebp, # 4		\ remove control word info
  mov   ebx, eax		\ set new TOS
  >FPU				\ restore FTOS
  fnext,
end-code

\ fixed SFP021
code f>d        \ F: f -- ; -- d ; convert float to double integer
\ *G Converts a float to a double integer.
\ ** Note that F>D truncates the number towards zero
\ ** according to the ANS specification.
\ ** See FR>D below.
  sub   ebp, # #12      	\ reserve space on stack
  fstcw word 0 [ebp]		\ save/preserve CW
  fstcw word 2 [ebp]
  or    word 0 [ebp], # $0C00	\ force truncation
  fldcw word 0 [ebp]
  fistp qword 4 [ebp]		\ store integer from FPU, -- dh dl oldtos
  fldcw word 2 [ebp]		\ restore rounding
  mov   eax, 8 [ebp]		\ get new TOS
  mov   8 [ebp], ebx		\ save old TOS
  add   ebp, # 4		\ remove control information
  mov   ebx, eax		\ set new TOS, -- oldtos dl dh
  >FPU				\ restore FTOS from deep stack
  fnext,
end-code

\ SFP022...
code fr>s	\ F: f -- ; -- n ; convert float to integer
\ *G Converts a float to a single integer using the current
\ ** rounding mode.
  sub   ebp, # 4        	\ reserve space on stack
  fistp dword 0 [ebp]		\ store integer from FPU
  mov   eax, 0 [ebp]		\ and put in reg
  mov   0 [ebp], ebx		\ preserve old TOS
  mov   ebx, eax		\ set new TOS
  >FPU				\ restore FTOS
  fnext,
end-code

code fr>d	\ F: f -- ; -- d ; convert float to double integer
\ *G Converts a float to a double integer using the current rounding mode.
  sub   ebp, # #8	      	\ reserve space on stack
  fistp qword 0 [ebp]		\ store integer from FPU, -- dh dl oldtos
  mov   eax, 4 [ebp]		\ get new TOS (dh)
  mov   4 [ebp], ebx		\ save old TOS
  mov   ebx, eax		\ set new TOS, -- oldtos dl dh
  >FPU				\ restore FTOS from deep stack
  fnext,
end-code
\ ...SFP022


\ **************
\ FP comparisons
\ **************

code fcmp1      \ F: f -- ; -- status
\ test the fp#, returning status result
  sub   ebp, # 4                \ save TOS
  mov   0 [ebp], ebx
  ftst                          \ test FTOS
  fstsw ax                      \ save status result
  fstp  st                      \ drop FPUstack
  and   ax, # $4100             \ mask out status bits
  movzx ebx, ax                 \ copy status to TOS
  >FPU                       \ reload FTOS
  fnext,
end-code

code fcmp2      \ F: f1 f2 -- ; -- status
\ test the top 2 floats, returning status
  sub   ebp, # 4                \ save TOS
  mov   0 [ebp], ebx
  >FPU				\ put second item on FPU
  fcompp                        \ compare and discard
  fstsw ax                      \ save status result
  and   ax, # $4100             \ mask out status bits
  movzx ebx, ax                 \ copy status to TOS
  >FPU				\ reload FTOS
  fnext,
end-code

: f0<           \ F: f1 -- ; -- t/f ; less than zero?
( *G Floating point 0<. N.B. result is on the Forth integer data stack. )
  fcmp1 $0100 =
;

: f0=           \ F: f1 -- ; -- t/f ; equal zero?
( *G Floating point 0=. N.B. result is on the Forth integer data stack. )
  fcmp1 $4000 =
;

: f0<>          \ F: f1 -- ; -- t/f ; not equal zero?
( *G Floating point 0<>. N.B. result is on the Forth integer data stack. )
  fcmp1 $4000 <>
;

: f0>           \ F: f1 -- ; -- t/f ; greater than zero?
( *G Floating point 0>. N.B. result is on the Forth integer data stack. )
  fcmp1 0=
;

: f<            \ F: f1 f2 -- ; -- t/f ; one less than the other?
( *G Floating point <. N.B. result is on the Forth integer data stack. )
  fcmp2 $0100 =
;

: f=            \ F: f1 f2 -- ; -- t/f ; equal each other?
( *G Floating point =. N.B. result is on the Forth integer data stack. )
  fcmp2 $4000 =
;

: f<>           \ F: f1 f2 -- ; -- t/f ; one not equal the other?
( *G Floating point <>. N.B. result is on the Forth integer data stack. )
  fcmp2 $4000 <>
;

: f>            \ F: f1 f2 -- ; -- t/f ; one less than the other?
( *G Floating point >. N.B. result is on the Forth integer data stack. )
  fcmp2 0=
;

: f<=           \ F: f1 f2 -- ; -- t/f ; one less or equal the other?
( *G Floating point <=. N.B. result is on the Forth integer data stack. )
  fcmp2
  dup $0100 =  swap $4000 = or
  0<>
;

: f>=           \ F: f1 f2 -- ; -- t/f ; one greater or equal the other?
( *G Floating point >=. N.B. result is on the Forth integer data stack. )
  fcmp2
  dup 0=  swap $4000 = or
  0<>
;

[-short-branches
: f~            \ F: f1 f2 f3 -- ; -- flag
( *G Approximation function. If f3 is positive, flag is true if abs[f1-f2] )
( ** is less than f3. If f3 is zero, flag is true if the f2 is exactly equal )
( ** to f1. If f3 is negative, flag is true if abs[f1-f2] less than )
( ** abs[f3*abs[f1+f2]]. )
  fdup f0> if                           \ if f3 > 0
    frot frot  f- fabs			\ -- f3 |f1-f2|
    f>                                  \ -- t/f		SFP007
  else
    fdup f0= if                         \ if f3 = 0
      fdrop  f=				\ f1=f2?		SFP007
    else                                \ if f3 < 0
      frot frot  fover fover		\ -- f3 f1 f2 f1 f2
      f- fabs                           \ -- f3 f1 f2 |f1-f2|
      frot frot                         \ -- f3 |f1-f2| f1 f2
      fabs fswap fabs                   \ -- f3 |f1-f2| |f1| |f2|
      f+ frot fabs f*                   \ -- |f1-f2| |f3|*(|f1|+|f2|)
      f<                                \ -- t/f
    endif
  endif
;
short-branches]


\ ******************************
\ Words dependent on FP compares
\ ******************************

: ?fnegate      \ F: f1 f2 -- f3
( *G Floating point NEGATE. )
  f0<
  if  fnegate  endif
;

: fmax          \ F: f1 f2 -- f3
( *G Floating point MAX. )
  fover fover f<
  if  fswap  endif
  fdrop
;

: fmin          \ F: f1 f2 -- f3
( *G Floating point MIN. )
  fover fover f>
  if  fswap  endif
  fdrop
;


\ ******************
\ FP logs and powers
\ ******************

code flog       \ F: f -- log(f)
( *G Floating point log base 10. )
  fldlg2
  fxch
  fyl2x
  fnext,
end-code

code fln        \ F: f -- ln(f)
( *G Floating point log base e. )
  fldln2
  fxch
  fyl2x
  fnext,
end-code

decimal
Lbl: 2**z       \ F: f -- 2^f ; SFP009
		\ This routine has to cope with range limitations

  push	edx
  sub	ebp, # #fpframe			\ make transient frame

\ make F absolute, and preserve previous sign
  sub	edx, edx			\ 0 for positive, +ve assumed
  ftst					\ test it
  fstsw	ax				\ save status, -ve if C3 C2 C0 = 001
  and 	ax, # $4100			\ extract C3,C2,C0
  cmp	ax, # $0100			\ -ve?
  z, if,
    dec   edx			        \ mark it so
    fabs				\ make absolute
  endif,

\ round FTOS down, preserving original FTOS
  fstcw word #cw [ebp]                  \ for later use
  fstcw word #rc [ebp]
  and   word #rc [ebp], # $f3ff         \ change the current control; was f3ff
  or    word #rc [ebp], # $0400         \ word to round down
  fldcw word #rc [ebp]
  fld   st                              \ copy number, and round one
  frndint                               \ N: -- f int
  fldcw word #cw [ebp]                  \ restore original control word

\ generate fraction from F and INT
  fxch					\ N: -- int f
  fld	 st(1)				\ N: -- int f int
  fsubrp st(1), st			\ N: -- int frac
  fchs					\ *** BUG in assembler FSUB, FSUBRP

\ generate 2^FRAC, preserving INT
  fld1
  fchs
  fxch					\ N: -- int -1 frac
  fscale				\ N: -- int -1 frac/2
  f2xm1					\ N: -- int -1 2^(frac/2)-1
  fxch
  fchs
  faddp st(1), st			\ N: -- int 2^(frac/2)
  fmul	st, st				\ N: -- int 2^frac

\ check sign of original F, applying as needed
  or	edx, edx
  nz, if,
    fld1
    fdivrp st(1), st			\ N: -- int 2^-frac
    fxch
    fchs
    fxch				\ N: -- -int 2^-frac
  endif,				\ N: -- int' 2^frac'

\ Apply the integer power to get 2^x, and clean up
  fscale				\ N: -- int' 2^(frac'+int')
  fxch
  fstp st				\ N: -- 2^(frac'+int'

  add   ebp, # #fpframe                 \ remove transient frame
  pop	edx

  ret
end-code

code 2**	\ F: f -- 2^f
( *G Floating point: returns 2^F. )
  call  2**z
  fnext,
end-code

code fexp       \ F: f -- e^f ; was called FE^X
( *G Floating point e^f. )
  fldl2e
  fmulp st(1), st
  call  2**z
  fnext,
end-code

: fexpm1	\ F: f -- (e^f)-1 ; 12.6.2.1516
\ *G Floating point log base (e^f)-1.
  %lg2e f* fdup %1 f< if
    [asm  f2xm1  asm]
  else
    2** %1 f-
  endif
;

code falog      \ F: f -- 10^f ; was called f10^f, new name: ans
( *G Floating point anti-log base 10. )
  fldl2t
  fmulp st(1), st
  call  2**z
  fnext,
end-code

code (f**)	\ F: f1 f2 -- f1^f2
\ *G Floating point returns f1 raised to the power f2.
\ ** N.B. no error checking is performed. If floating point
\ ** execeptions are masked, which is the default condition,
\ ** the system will return a NaN for f1<0.
  >FPU
  fyl2x
  call  2**z
  fnext,
end-code

: f**		\ F: f1 f2 -- f1^f2 ; was called fx^y ; SFP009
\ *G Floating point: returns f1 raised to the power f2.
\ ** If f1<=0e0, 0e0 is returned. This behaviour is required
\ ** by the Forth Scientific Library.
  fover f0> if
    (f**)  exit
  endif
  fdrop fdrop %0
;


\ ********
\ Rounding
\ ********

code [fround]   \ mode -- ; F: f -- f' ; round floating point value
  sub   ebp, # #fpframe         \ allow space for results
  fstcw #cw [ebp]               \ get control word
  mov   cx, #cw [ebp]           \ and preserve it
  and   word #cw [ebp], # $F3FF \ strip rounding control
  or    word #cw [ebp], bx      \ or in new mode
  fldcw #cw [ebp]               \ put new rounding control back
  frndint                       \ rounding operation
  mov   #cw [ebp], cx           \ restore previous control word
  fldcw #cw [ebp]
  add   ebp, # #fpframe         \ discard control word
  mov   ebx, 0 [ebp]            \ restore TOS
  add   ebp, # 4
  fnext,
end-code

: fround        \ F: f1 -- f1'
( *G Round the number to nearest or even. )
  0 [fround]
;

: fint          \ F: f1 -- f1'
( *G Chop the number towards zero. )
  $0C00 [fround]
;

: floor         \ F: f1 -- f1'
( *G Floored round towards -infinity. )
  $0400 [fround]
;

: roundup       \ F: f1 -- f1'
( *G Round towards +infinity. )
  $0800 [fround]
;

: rounded	\ -- ; set NDP to round to nearest
( *G Set NDP to round to nearest for all operations other than )
( ** FINT FLOOR and ROUNDUP. )
  cw@ $F3FF and ( $0 or ) cw!
;
rounded

: floored	\ -- ; set NDP to floor
( *G Set NDP to round to floor for all operations other than )
( ** FROUND FINT and ROUNDUP. )
  cw@ $F3FF and $400 or cw!
;

: roundedup	\ -- ; set NDP to round up
( *G Set NDP to round up for all operations other than )
( ** FROUND FINT and FLOOR. )
  cw@ $F3FF and $800 or cw!
;

: truncated	\ -- ; set NDP to chop to 0
( *G Set NDP to chop to 0 for all operations other than )
( ** FROUND FLOOR and ROUNDUP. )
  cw@ $F3FF and $C00 or cw!
;


\ ****************************
\ Number conversion primitives
\ ****************************

\ *****************************
\ Floating point numeric output
\ *****************************

decimal

((
>BCD and >ASCII
===============
The following print control words are destructive, both to the
86 stack and the 87.  Too many numbers on the 87 stack will
result in an 87 stack overflow, which is not trapped out.
The 86 stack and the text buffer are serially written to and
then read from, altering the stack size accordingly, and finally
reset to point to the next free location after the 86 has placed
the 'real' floating point number at S0.
All conversions are messy, going through i) ascii to bcd, ii)
bcd to integer, iii) integer to floating, and iv) integer to
exponent.  Implementation of these steps is necessary however,
and attention to detail is required.
The result will be accurate to 2 digits in the 15th decimal
place.

The print command makes extensive use of the text buffer to hold
both the value and its exponent, and intermediate results from
the bcd to ascii conversion.  It is destructive of ANYTHING that
was in the text buffer.

>BCD
====
This subroutine converts the floating point number on the 87
stack to a bcd string and power of ten value in the 8086 text
buffer area.
Unnormal, denormal, and pseudo zeros are not trapped out, but
do not seem to significantly affect the conversion.
The full 18 digit capability of the (8087) data register is used
but the final ascii string will only be 17 digits long.  The
generated ascii string will always have a leading "+" or "-".
The decimal place will always be placed after the first digit,
although the input conversion routines allow free formatting.
It is assumed that the text buffer is longer than 35 bytes, and
that the 8086 stack has at least five free entries.
Do not alter any of the next few routines, they are quite
complex!!!

The number is converted within the following areas

PAD      = ascii_string (20 bytes)
PAD + 20 = scaling factor for power_of_ten
PAD + 24 = power_of_ten of exponent using FLOG
PAD + 28 = power_of_ten value for display
PAD + 34 = power_of_two
PAD + 42 = fraction or bcd string area (18 bytes)
PAD + 60 = integer or real number flag

TOS is used to carry the sign from one routine to the other
))

hex
create ^1e17  00 c, 00 c, 00 c, C5 c, 2E c, BC c, A2 c, B1 c,  37 c, 40 c,
decimal

code >bcd               \ pad -- flag ; F: f -- ; assumes number already on 87 stack
                        \ PAD is in TOS (EBX), and is a valid address
		\ ROUNDED has already been imposed!
  cld                                   \ increment all further instructions
  sub   ebp, # #fpframe                 \ build frame

\ Fill PAD with 20 bytes of ASCII `0'
  mov   edx, edi                        \ preserve EDI (Forth reg)
  mov   al, # char 0
  mov   ecx, # 20
  mov   edi, ebx                        \ es:di==pad
  rep stosb                             \ initialise the pad area for type; al=48, ecx=20
  mov   edi, edx                        \ restore EDI

  xor   eax, eax                        \ setting `real/int' and `power_10' to zero
  mov   dword 60 [ebx], eax             \ assume real unless told otherwise; ds:di==pad
  mov   dword 28 [ebx], eax             \ zero the power of ten value

\ Test and save the sign of the top item of the FP stack
  ftst                                  \ determine sign of top 87 stack word
  fstsw word #sw [ebp]                  \ #sw [BP] = sw
  and   word #sw [ebp], # $4100         \ 04100h == C3 and C0
  test  word #sw [ebp], # $0100         \ 0100h == look at C0, set if <0.0
  nz, if,
    mov   edx, # 1                      \ if -ve,
  else,
    mov   edx, # 0                      \ if +ve, corresponding values used
  endif,

  test  word #sw [ebp], # $4000         \ skip the power code routine if
  jne   l$3                             \ the number was zero, C3 is FZ bit
  fabs                                  \ work only with positive numbers
\ n: f --
  fxtract                               \ separate the significand/exponent
  fstp  tbyte 42 [ebx]                  \ save the fraction temporarily; ebx==pad
  fist  dword 34 [ebx]                  \ save the exponent temporarily, then
  fldlg2                                \  calculate the proper power of ten.
  fmulp st(1), st                       \  exponent * lg2 base 10
  fistp dword 28 [ebx]                  \ store the power_of_ten value for display
  mov   dword eax, 28 [ebx]             \ immediately retrieve it again,
  sub   eax, # 17                       \ and calculate the scaling factor to
					\ get the number to fit into bcd limit of
			 		\ sign+17 characters
  mov   dword 20 [ebx], eax             \ set initial power_of_ten,
  ja    l$6                             \ jump if eax u> 17d

  fild  dword 34 [ebx]                  \ load power_of_two, and
  fld   tbyte 42 [ebx]                  \ load the fraction
  fscale                                \ then set up the number
  fxch                                  \ fswap
  fstp  st                              \ fdrop
  fld   st                              \ fdup
  jmp   l$5
l$3:                                    \ branch island for -128..+127 bytes
  mov   dword 60 [ebx], # 1             \ force result to integer
  jmp   l$4
l$5:

\ determine whether result can be displayed as an integer, or
\ must be displayed in FP format.
  frndint                               \ convert FTOS to integer
  fcomp  st(1)                          \ compare with the real version
  fstsw  word #sw [ebp]                 \ the equality of the number to; ebx==sw
  test   word #sw [ebp], # $4000        \ its integer equivalent
  nz, if,
    mov  dword 60 [ebx], # 1            \ yes, an integer between 1 and 1e18
    jmp  l$4                            \ go to end of proc. via l$11 to l$10
  endif,

  mov dword 60 [ebx], # 0               \ no, non-integer...need to scale to range
  fstp  st                              \ fdrop
l$6:            \ this label is only for numbers needing FP format display
                                        \ EAX holds the scaled exponent
  fild  dword 20 [ebx]                  \ load power_of ten value
  fchs                                  \ negate FTOS
  fldl2t                                \ load log 10 base 2 onto 87 stack
  fmulp st(1), st                       \ power_of_ten * log 10 base 2
  call  2**z                            \ calculate scaling factor and
  fxtract                               \ get scaled up number, power 2
  fld 	tbyte 42 [ebx]                  \ load the fraction part of the number
  fmulp st(1), st
\ and combine to within the range allowed by the bcd format,
\ to within an order of magnitude.
  jmp   l$2
l$4:                                    \ this jump dest is to get round the branch size
  jmp   l$10                            \ ...ditto...
l$2:
  fild  dword 34 [ebx]                  \ load the power_of_two value
  faddp st(2), st                       \ combine the powers of two, to form
  fscale                                \ the full value of the number without
  fxch                                  \ affecting the value of the exponent.
  fstp  st
  frndint                       	\ convert to integer
  fld	tbyte ptr ^1e17			\ form 10^17
  fcomp st(1)                           \ test the adjusted number to see if
                                        \ the error is one order of magintude
  fstsw word #sw [ebp]                  \ too large to fit into bcd field
  test  word #sw [ebp], # $4100         \ if c0 = 0 then number is >= to
  z, if,                                \ to the lower boundary
    dec  dword 28 [ebx]                 \ adjust power_of_ten value
  endif,

l$10:           \ all numbers pass this way
  fbstp 42 [ebx]                        \ store the integer mantissa in pad
  					\ performs integer rounding

  mov   ebx, edx                        \ flag required by next word

  add   ebp, # #fpframe                 \ release frame

  >FPU                                  \ restore FTOS
  fnext,
end-code

hex

code >ascii     \ pad flag -- length
\ Save the Forth registers
  push  esi
  push  edi

  mov   edi, 0 [ebp]                    \ DS:DI=pad
  add   ebp, # 4

  mov   edx, ebx                        \ EDX=flag

  push  ebp                             \ preserve data stack pointer!

  mov   ebp, edi                        \ ebp==pad, PAD must be in stack segment!
  mov   esi, # $08                      \ initial bcd size for index count
  mov   ecx, # $0F04                    \ shift count and mask
  mov   ebx, # 01                       \ initial field size for ascii string
  cld                                   \ autoincrement upwards in memory
  mov   al, # char +                    \ set "sign" field to "+"
  test  dl, # 01                        \ look for a negative number, and if so
  jz    l$1                             \  then change the
    mov   al, # char -                  \  "sign" field to "-"
l$1:
  stosb                                 \ bump sign character out to pad
  and   dl, # $07F                      \ turn off sign bit
l$2:            \ now remove surplus zeros from field
  mov   byte ah, $02A [ebp] [esi]       \ retrieve a bcd byte from PAD
  mov   al, ah                          \ copy the byte's value
  shr   al, cl                          \ get high order digit, and test the
  and   al, ch                          \ high order bit to see if leading
  jnz   l$4                             \ zero found, if not, print number
  mov   al, ah                          \ get bcd byte again
  and   al, ch                          \ test the low order digit, and see
  jnz   l$5                             \ if that was zero, if not, exit
  dec   esi                             \ dec the bcd index correspondingly
  jns   l$2                             \ if the significand has all zeros
  mov   al, # char 0                    \ then only print an initial zero
  stosb                                 \ bump string out to pad
  inc   ebx                             \ and exit the session
  jmp l$6
l$3:
  mov   byte ah, $02A [ebp] [esi]       \ retrieve a bcd byte from PAD
  mov   al, ah
  shr   al, cl                          \ get high order digit
l$4:
  add   al, # char 0                    \ convert to ascii by bumping 48
  stosb                                 \ bump string out to pad
  mov   al, ah                          \ get low order digit
  and   al, ch
  inc   ebx                             \ increase the field size counter
l$5:
  add   al, # char 0                    \ convert to ascii by bumping 48
  stosb                                 \ bump string out to pad
  inc   ebx                             \ increase the field size counter
  dec   esi                             \ try next bcd byte value
  jns l$3

l$6:            \ conversion is now complete
  pop   ebp                             \ Restore the Forth registers
  pop   edi
  pop   esi

  fnext,
end-code

: >exp		\ -- n ; F: f -- f ; EDA024 SFP025
\ Return decimal exponent as integer n
  { | temp[ 8 ] -- }
  fdup f0= if
    0
  else
    fdup fabs flog
    temp[ df!  temp[ df@
    floor f>s
  endif
;

: f>ascii	\ buff -- length ; F: f -- ; SFP009 ; EDA024
  cw@ >r  rounded >exp
  over #24 + !
  dup >bcd >ascii
  r> cw!
;


\ ****************
\ Compiler support
\ ****************

decimal

code flit       \ F: -- f ; takes floating point number inline
  pop   edx                     \ get return address
  FPU>                          \ save current FTOS
  fld   fword 0 [edx]           \ load new inline value
  add   edx, # FPCELL
  jmp   edx
end-code

defer fliteral	\ F: f -- ; F: -- f
\ *G Compiles a float as a literal into the current definition.
\ ** At execution time, a float is returned. For example,
\ ** [ %PI F2* ] FLITERAL will compile 2PI as a floating point
\ ** literal. Note that FLITERAL is immediate, whereas (RLITERAL)
\ ** below is not.
  immediate

: (rliteral)	\ F: f -- ; F: -- f
( *G Compiles a float as a literal into the current definition. )
( ** At execution time, a float is returned. For example, )
( ** [ %PI F2* ] FLITERAL will compile 2PI as a floating point )
( ** literal. The default action of FLITERAL. Note that FLITERAL )
( ** is immediate, whereas [FLITERAL] is not. )
\  postpone flit  f,  discard-sinline
  [ also h-c ]
  GenTokens? if				\ SFP027...
    fdup genFlitToken  0 -> GenTokens?
    postpone flit  f,  1 -> GenTokens?
  else
    postpone flit  f,
  endif					\ ...SFP027
  [ previous ]
;
assign (rliteral) to-do fliteral

: #fliteral              \ n -- F: real --
  drop  postpone fliteral
;

((
\ Alternate optimised version, but if you really want speed use
\ NDP387.FTH instead. Left here for design notes purposes.
also assembler  also asm-access
: o_flit,	\ F: f -- ; -- f ; compile floating point literal
  mov   eax, FSP-OFFSET [esi]   \ get FP stack pointer
  lea	eax, -FPCELL [eax]	\ update stack pointer
  fstp  fword 0 [eax]           \ store and pop
  mov   FSP-OFFSET [esi], eax
  jmp   here 2+  FPCELL +	\ skip inline literal
  asm-end asm-off
  f,
  asm-start asm-on
  fld	fword ptr here FPCELL -
  asm-end asm-off
;
previous previous

assign o_flit, to-do fliteral
))


\ ***************
\ FP trigonometry - 387 and up required
\ ***************

hex

code ftan       \ F: f -- tan(f)
( *G Floating point tangent. )
  fptan
  fstp  st
  fnext,
end-code

code fatan      \ F: f -- atan(f)
( *G Floating point arctangent. )
  fld1
  fpatan
  fnext,
end-code

code fsin       \ F: f -- sin(f)
( *G Floating point sine. )
  fsin
  fnext,
end-code

code fasin      \ F: f -- asin(f)
\ *G Floating point arcsine.
  fld    st
  fld    st
  fmulp  st(1), st
  fld1
  fsubrp st(1), st
  fsqrt
  fpatan
  fnext,
end-code

code fcos       \ F: f -- cos(f)
( *G Floating point cosine. )
  fcos
  fnext,
end-code

code facos      \ F: f -- acos(f)
\ *G Floating point arctangent.
  fld    st
  fld    st
  fmulp  st(1), st
  fld1
  fsubrp st(1), st
  fsqrt
  fxch   st(1)
  fpatan
  fnext,
end-code

code fsincos    \ F: f -- sin(f) cos(f)
( *G Returns sine and cosine values of f. )
  fsincos
  fxch
  FPU>
  fnext,
end-code

code fatan2	\ F: f1 f2 -- atan(f1/f2)
\ *G Floating point arctangent with prior division.
  >FPU					\ f1 -> st(0)
  fpatan
  fnext,
end-code

: deg>rad	\ F: fdeg -- frad
( *G Converts a value in degrees to radians. )
  %pi f*  #180 s>f f/
;

: rad>deg               \ -- ;
( *G Converts a value in radians to degrees. )
  #180 s>f f*  %pi f/
;

code freduce    \ F: f1 -- f2 ; reduce value to range 0..2pi
( *G Reduce f1 to be in the range 0 <= f2 < 2PI. )
  begin,
    fldpi                       \ constant of 2pi
    fadd   st, st
    fxch
    fprem                       \ reduce
    fstsw  ax                   \ get status
    fxch			\ remove 2pi
    fstp   st
    and    ax, # $0400          \ inspect C2 bit
  z, until,                     \ until C2 clear
  fnext,
end-code


\ ******************************************************
\ Extra trig functions - COSEC SEC COTAN and hyberbolics
\ ******************************************************

: fcosec        \ F: f -- cosec(f)
( *G Floating point cosecant. )
  fsin 1/f  ;

: fsec          \ F: f -- sec(f)
( *G Floating point secant. )
  fcos 1/f  ;

: fcotan        \ f: f -- cot(f)
( *G Floating point cotangent. )
  ftan 1/f  ;

: fsinh         \ F: f -- sinh(f) ; (e^x - 1/e^x)/2
( *G Floating point hyberbolic sine. )
  fexp  fdup 1/f f-  f2/  ;

: fcosh         \ F: f -- cosh(f) ; (e^x + 1/e^x)/2
( *G Floating point hyberbolic cosine. )
  fexp  fdup 1/f f+  f2/  ;

: ftanh         \ F: f -- tanh(f) ; (e^x - 1/e^x)/(e^x + 1/e^x)
( *G Floating point hyberbolic tangent. )
  fdup 1/f  fover fover f-
  frot frot f+  f/
;

: fasinh        \ F: f -- asinh(f) ; ln(f+sqrt(1+f*f))
( *G Floating point hyberbolic arcsine. )
  fdup
  fdup f*  %1 f+  fsqrt
  f+  fln  ;

: facosh        \ F: f -- acosh(f) ; ln(f+sqrt(f*f-1))
( *G Floating point hyberbolic arccosine. )
  fdup
  fdup f*  %1 f-  fsqrt
  f+  fln  ;

: fatanh        \ F: f -- atanh(f) ; ln((1+f)/(1-f))/2
( *G Floating point hyberbolic arctangent. )
  %1 fover f+  %1 frot f-  f/
  fln  f2/  ;



\ *****************
\ Number conversion
\ *****************

((
Converting from ASCII to floating point is not so hard as before
since there are some FORTH words for some of the jobs.  It still
consists of four basic steps: i) obtain the significand and its
decimal maginitude, ii) get the exponent, iii) convert ascii to
integer, and then iv) correct the power value.

The input format takes the following form:

[-] [digit(s)] [.] [digit(s)] [E|e] [+|-] [digit(s)]
))

decimal

((
: tfv,		\ -- ; lay a 10 byte float from ten hex values
  base @ >r  hex
  #10 0 do
    bl word integer? 1 <> abort" bad byte" c,
  loop
  r> base !
;

: >fcon		\ -- ; F: -- f ; convert previous word to FCONSTANT
  does>
    tf@
;

: >farr		\ -- ; n -- addr ; convert previous word to FCONARRAY
  does>
    swap #10 * + tf@			\ make address, get data
;


create %1e256   tfv, 06 DE F9 9D FB EB 7E AA  51 43  >fcon	\ 8C DE F9 9D  FB EB 7E AA--51 43
create %1E-1    tfv, CD CC CC CC CC CC CC CC  FB 3F  >fcon
create %1E-10   tfv, BF D5 ED BD CE FE E6 DB  DD 3F  >fcon
create %1e-256  tfv, 3C 19 7A 63 25 43 31 C0  AC 3C  >fcon	\ 3C 19 7A 63  25 43 31 C0--AC 3C

create POWERS-OF-10E1	\ 1e0 .. 1e15
  tfv, 00 00 00 00 00 00 00 80  FF 3F		tfv, 00 00 00 00 00 00 00 A0  02 40
  tfv, 00 00 00 00 00 00 00 C8  05 40		tfv, 00 00 00 00 00 00 00 FA  08 40
  tfv, 00 00 00 00 00 00 40 9C  0C 40		tfv, 00 00 00 00 00 00 50 C3  0F 40
  tfv, 00 00 00 00 00 00 24 F4  12 40	        tfv, 00 00 00 00 00 80 96 98  16 40
  tfv, 00 00 00 00 00 20 BC BE  19 40		tfv, 00 00 00 00 00 28 6B EE  1C 40
  tfv, 00 00 00 00 00 F9 02 95  20 40		tfv, 00 00 00 00 40 B7 43 BA  23 40
  tfv, 00 00 00 00 10 A5 D4 E8  26 40	        tfv, 00 00 00 00 2A E7 84 91  2A 40
  tfv, 00 00 00 80 F4 20 E6 B5  2D 40	        tfv, 00 00 00 A0 31 A9 5F E3  30 40
>farr

create POWERS-OF-10E-1	\ 1e0 .. 1e-15
  tfv, 00 00 00 00 00 00 00 80  FF 3F		tfv, CD CC CC CC CC CC CC CC  FB 3F	\   0  -1
  tfv, 0A D7 A3 70 3D 0A D7 A3  F8 3F		tfv, 3B DF 4F 8D 97 6E 12 83  F5 3F	\  -2  -3
  tfv, 2C 65 19 E2 58 17 B7 D1  F1 3F		tfv, 23 84 47 1B 47 AC C5 A7  EE 3F	\  -4  -5
  tfv, B6 69 6C AF 05 BD 37 86  EB 3F	        tfv, BC 42 7A E5 D5 94 BF D6  E7 3F	\  -6  -7
  tfv, FD CE 61 84 11 77 CC AB  E4 3F		tfv, 97 A5 B4 36 41 5F 70 89  E1 3F	\  -8  -9
  tfv, BF D5 ED BD CE FE E6 DB  DD 3F		tfv, FF AA 24 CB 0B FF EB AF  DA 3F	\ -10 -11
  tfv, CC 88 50 6F 09 CC BC 8C  D7 3F		tfv, 13 0E B4 4B 42 13 2E E1  D3 3F	\ -12 -13
  tfv, 0F D8 5C 09 35 DC 24 B4  D0 3F		tfv, D9 AC B0 3A F7 7C 1D 90  CD 3F	\ -14 -15
>farr

create POWERS-OF-10E16	\ 1e0 .. 1e256
  tfv, 00 00 00 00 00 00 00 80  FF 3F		tfv, 00 00 00 04 BF C9 1B 8E  34 40	\   0  16
  tfv, 9E B5 70 2B A8 AD C5 9D  69 40	        tfv, D7 95 43 0E 05 8D 29 AF  9E 40	\  32  48
  tfv, D5 A6 CF FF 49 1F 78 C2  D3 40		tfv, FB F7 DA 87 8F 7A E7 D7  08 41	\  64  80
  tfv, A4 14 9B C5 16 AB B3 EF  3D 41	        tfv, 9C 32 23 99 C0 AD 0F 85  73 41	\  96 112
  tfv, DF 8C E9 80 C9 47 BA 93  A8 41		tfv, E6 A6 D3 A8 C5 B9 02 A4  DD 41	\ 128 144
  tfv, A9 17 E6 7F 2B A1 16 B6  12 42	        tfv, 95 BF 9B 85 91 A2 28 CA  47 42	\ 160 176
  tfv, 6D 55 27 39 8D F7 70 E0  7C 42		tfv, 45 61 82 37 35 0C 2E F9  B1 42	\ 192 208
  tfv, 2B C9 3C E3 FF 96 52 8A  E7 42		tfv, 69 17 BF D6 F3 A6 91 99  1C 43	\ 224 240
  tfv, 8C DE F9 9D FB EB 7E AA  51 43	 						\ 256
>farr

create POWERS-OF-10E-16	\ 1e0 .. 1e-256
  tfv, 00 00 00 00 00 00 00 80  FF 3F		tfv, 5B E1 4D C4 BE 94 95 E6  C9 3F	\    0  -16
  tfv, BA 94 39 45 AD 1E B1 CF  94 3F	        tfv, 5A C1 7E B1 53 7C 12 BB  5F 3F	\  -32  -48
  tfv, A5 E9 39 A5 27 EA 7F A8  2A 3F		tfv, A6 19 09 6B BA 60 C5 97  F5 3E	\  -64  -80
  tfv, 9B 53 75 FD F7 02 B4 88  C0 3E		tfv, 80 D3 65 F0 BC 35 43 F6  8A 3E	\  -96 -112
  tfv, A1 E4 BC 64 7C 46 D0 DD  55 3E		tfv, CB 82 53 7C 6E BA CA C7  20 3E	\ -128 -144
  tfv, 94 A0 73 DB 93 E0 F4 B3  EB 3D		tfv, 2F 00 CB 38 DB 27 17 A2  B6 3D	\ -160 -176
  tfv, 6D 55 27 39 8D F7 70 E0  7C 42		tfv, 60 BC A4 3D A9 DE 80 83  4C 3D	\ -192 -208
  tfv, C5 4E 31 4A EC 3C E5 EC  16 3D		tfv, D2 E1 32 CF CD 5F 60 D5  E1 3C	\ -224 -240
  tfv, 3C 19 7A 63 25 43 31 C0  AC 3C							\ -256
>farr

[-sin
: RAISE_POWER	\ mant exp -- mant' exp'
  DUP >R  #15 AND ?DUP
  IF  POWERS-OF-10E1 F*  THEN
  R@ 4 rshift  #15 AND ?DUP
  IF  #16 MIN POWERS-OF-10E16 F*  THEN
  R> 8 rshift  0
  ?DO  %1e256 F*  LOOP
;

: SINK_FRACTION	\ F: f -- f' ; exp --
  ABS dup >R  #15 AND ?DUP
  IF  POWERS-OF-10E-1 F*  THEN
  R@ 4 rshift #15 AND ?DUP
  IF  POWERS-OF-10E-16 F*  THEN
  R> 8 rshift  0
  ?DO  %1e-256 F*  LOOP
;
+sin]

: *10^X		\ f: f -- f' ; exp -- ; f' = f *10^exp
  DUP 0< IF
    SINK_FRACTION
  ELSE
    RAISE_POWER
  THEN
;

: 10**n		\ n -- ; F: -- f
  %1 *10^x
;
))

create powers10	\ -- addr ; table of powers of 10
                    1, swap , ,		\ in Intel order (low, high)
		   10, swap , ,		\ 10^1
		  100, swap , ,		\ 10^2
		 1000, swap , ,		\ 10^3
		10000, swap , ,		\ 10^4
	       100000, swap , ,		\ 10^5
	      1,000000 swap , ,		\ 10^6
	     10,000000 swap , ,		\ 10^7
	    100,000000 swap , ,		\ 10^8
	   1000,000000 swap , ,		\ 10^9
	  10000,000000 swap , ,		\ 10^10
	 100000,000000 swap , ,		\ 10^11
	1,000000000000 swap , ,		\ 10^12
       10,000000000000 swap , ,		\ 10^13
      100,000000000000 swap , ,		\ 10^14
     1000,000000000000 swap , ,		\ 10^15
    10000,000000000000 swap , ,		\ 10^16
   100000,000000000000 swap , ,		\ 10^17
  1000000,000000000000 swap , ,		\ 10^18

code get10^n	\ n -- ; -- f ; 0<=n<=18
  FPU>				\ save previous FTOS
  mov	eax, # powers10
  fild  qword 0 [eax] [ebx*8]	\ load integer from table
  mov	ebx, 0 [ebp]		\ restore TOS
  lea	ebp, 4 [ebp]
  fnext,
end-code

: (10**n)	\ +n -- ; -- f
  dup #18 > if
    #10 s>f  s>f f**  exit
  endif
  get10^n
;

: 10**n		\ n -- ; -- f
  dup abs (10**n)  0<
  if  1/f  endif
;

: exp-char?	\ char -- flag ; true if character is exponent indicator
  case  upc
    [char] D of  true  endof
    [char] E of  true  endof
                 drop false
  end-case
;

[-short-branches
: >FLOAT	\ c-addr u -- flag ; F: -- f | --
( *G Try to convert the string at c-addr/u to a floating point number.  )
( ** If conversion is successful, flag is returned true, and a floating )
( ** number is returned on the float stack, otherwise flag is returned  )
( ** false and the float stack is unchanged.                            )
  { c-addr u | dlo dhi u' fsign flag nshift obase -- flag }

  u 0= IF false exit THEN               \ must have at least one significand digit

  base @ -> obase  decimal		\ preserve and force base ; SFP028

  false -> flag
  0 -> nshift

  c-addr c@ [char] + = if               \ Ignore preceeding +
    c-addr 1 chars + -> c-addr
    u 1 chars - -> u
  then

  c-addr c@ [char] - =                  \ check for preceeding -
  dup -> fsign  IF
    c-addr 1+ -> c-addr
    u 1- -> u
  THEN

  \ convert mantissa (first set of digits)
  0 0 c-addr u >number -> u' -> c-addr -> dhi -> dlo

  u' 0> IF
    \ convert optional number after decimal point
    c-addr c@ fp-char @ = IF
      dlo dhi c-addr 1+ u' 1- dup -> nshift >number
      dup nshift - -> nshift
      -> u' -> c-addr -> dhi -> dlo
    THEN
    \ convert exponent
    u' 0> IF
      c-addr c@ exp-char? IF
        u' 1 > if
          c-addr 1+ u' 1- (integer?)
        else
          0 1
        then
        IF
          nshift + -> nshift
          true -> flag
        THEN
      THEN
    ELSE
      true -> flag
    THEN
  ELSE
    u u' <> -> flag
  THEN

  \ convert double precision int to float
  flag IF
    cw@ >r  rounded
    dlo dhi d>f
    nshift 10**n f*
    fsign IF  fnegate  THEN
    r> cw!
  THEN

  flag  obase base !
;
short-branches]


\ --------------------------
\ f. sub words - places, etc
\ --------------------------

decimal

variable #places
  8 #places !

: precision     \ -- u
\ *G Returns the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
  #places @
;

: set-precision \ u --
\ *G Sets the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
  0 max  #fdigits min  #places !
;

8 set-precision

: places        \ u --
\ *G Sets the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
\ ** The ANS version of this word is *\fo{SET-PRECISION}, which
\ ** should be used in new code.
  set-precision
;

: real.         \ c-addr flag -- ; show the significand
  over c@ [char] - = or if		\ decide whether to show sign
    dup c@ emit
  endif
  dup 1 + c@ emit
  fp-char @ emit			\ SFP006
  dup 2+  precision 1- 0 max type
  dup 28 + @ if				\ SFP019
    ." e"  28 + @ .
  else
    drop  space
  endif
;

\ The next three words assume that the value to be tested has
\ been placed in a 10-byte real whose address is passed to them.

\ SFP012...
: fsign?	\ addr -- flag ; true if negative
  8 + w@ $8000 and 0<>
;

: infinity?	\ addr -- flag ; true if infinite
  dup 8 + w@ $7FFF and $7FFF =
  swap 2@ $80000000 $00000000 d= and
;

: NaN?		\ addr -- flag ; true if not a number
  dup 8 + w@ $7FFF and $7FFF =
  swap 2@ $80000000 $00000000 d= 0= and
;
\ ...SFP012

: BadFloat?	\ F: f -- ; -- caddr u true | false
\ *G If the float is a NaN or Infinite, return a string such as
\ ** "+NaN" and true, otherwise just return false (0).
  { | checkit[ #10 ] -- }
  checkit[ tf!				\ save for test
  checkit[ NaN? if
    checkit[ fsign?
    if  s" -NaN"  else  s" +NaN"  endif
    true  exit
  endif
  checkit[ infinity? if
    checkit[ fsign?
    if  s" -Inf"  else  s" +Inf"  endif
    true  exit
  endif
  false
;

: represent	\ c-addr u -- n flag1 flag2 ; F: f --
( *G Assume that the floating number is of the form +/-0.xxxxEyy. )
( ** Round the significand xxxxx to u significant digits and place its )
( ** representation at c-addr. If u is zero round the fractional )
( ** significand to a whole number. flag2 is true if the results are )
( ** valid. n is the signed integer version of yy and flag1 is true if )
( ** f is negative. In this implementation all errors are handled by )
( ** exceptions, and so flag2 is always true, except for NaNs and )
( ** Infinites. The number of characters at c-addr is the greater of )
( ** u or MAX-FLOAT-DIGITS. )
  { caddr u | checkit[ #10 ] sign exp -- n flag1 flag2 }
  caddr u #fdigits max blank		\ set to spaces for text messages
  fdup BadFloat? if
   over c@ [char] - = >r		\ sign of result
   1 /string				\ strip sign
   caddr swap cmove			\ set o/p to "NaN" or "Inf"
   3 r> 0				\ -- 3 sign 0
   fdrop exit
  endif
  caddr u
  2dup #fdigits max [char] 0 fill	\ prefill with '0' chars
  #fdigits min  2>r
  >exp 1+ to exp
  r@ exp - 10**n  f*
  fround  fdup f0< to sign
  fabs f>d  2dup <# #s #>
  dup r@ - addr exp +!
  2r>  rot min  1 max  cmove
  d0= if  1 false  else  exp sign  endif
  true					\ exceptions handled by CPU and CATCH/THROW
;

((
: represent	\ c-addr u -- n flag1 flag2 ; F: f --
 ( *G Assume that the floating number is of the form +/-.xxxxEyy. )
 ( ** Place the significand xxxxx at c-addr rounded to u significant )
 ( ** digits. If u is zero place one digit at c-addr representing )
 ( ** the fractional significand r rounded to a whole number. )
 ( ** Rounding is performed according to FROUND. )
 ( ** Return n the signed integer version of yy. Return flag1 true )
 ( ** if f is negative, and return flag2 true if the results are )
 ( ** valid. In this implementation all errors are handled by )
 ( ** exceptions, and so flag2 is always true, except for NaNs and )
 ( ** Infinites. )
\ SFP012...
  { caddr u | checkit[ #10 ] sign exp -- n flag1 flag2 }
  caddr u blank				\ set to spaces for text messages
  fdup checkit[ tf!			\ save for test
  checkit[ NaN? if
    s" NaN" caddr swap u min cmove	\ set o/p to "NAN"
    3 u min checkit[ fsign? 0		\ return 3 sign 0 for NAN
    fdrop  exit
  endif
  checkit[ infinity? if
    s" Inf" caddr swap u min cmove	\ set o/p to "INF"
    3 u min checkit[ fsign? 0		\ return 3 flag 0 for INF
    fdrop  exit
  endif
\ ...SFP012
  caddr u  2dup [char] 0 fill		\ prefill with '0' chars
  #18 min  2>r
  fdup f0< to sign
  fabs >exp 1+ to exp
  r@ exp - 10**n  f*
  fround f>d  2dup <# #s #>
  dup r@ - addr exp +!
  2r>  rot min  1 max  cmove
  d0= if  1 false  else  exp sign  endif
  true					\ exceptions handled by CPU and CATCH/THROW
;
))

: fe.		\ F: f -- ; display in engineering format
( *G Display f in engineering notation, in which the exponent is always a power )
( ** of three, and the significand is always in the range 1.xxx to 999.xxx. )
  fdup BadFloat?
  if  type  fdrop exit  endif

  base @ decimal
  pad dup f>ascii			\ -- pad len ; generate ASCII text	SFP009
  precision 1+ min			\ allowing for sign and PRECISION
\ Calculate modified exponent
  pad 28 + @ s>d 3 fm/mod 3 * >r	\ -- pad len step
\  pad 24 + @ s>d 3 fm/mod 3 * >r	\ -- pad len step ; SFP019
\ Display significand
  2+ 0 do
    over c@ emit  1 /string		\ -- pad' len' ; before decimal point
  loop
  fp-char @ emit			\ SFP006
  0 max dup
  if  type  else  2drop  endif		\ -- ; after decimal point
\ display exponent
  r>
  [char] E emit
  dup 0< if  ." -"  else  ." +"  endif
  abs .
  base !
;

: fs.		\ F: f -- ; display in scientific format
( *G Display f in scientific notation, with one digit before the )
( ** decimal point. )
  fdup BadFloat?
  if  type  fdrop exit  endif

  base @ decimal
  pad f>ascii drop			\ generate ASCII text		SFP009
  pad true real.			\ sign always displayed
  base !
;

[-short-branches
: f.            \ F: f -- ; display number in FTOS
( *G Display f as a float in fixed point notation with a trailing space. )
  fdup BadFloat?
  if  type  fdrop exit  endif

  base @ decimal
  pad f>ascii				\ SFP009
  pad 60 + c@ if			\ true if integer
    pad swap type space
  else
    drop  pad false real.
  endif
  base !
;
short-branches]

: f?		\ addr -- ; displays contents of FVARIABLE
( *G Displays the contents of the given FVARIABLE. )
  f@ f. ;

: f.s           \ F: i*f -- i*f
( *G Display the contents of the floating point stack. )
  fdepth 0< -400 ?throw
  fdepth if
    cr ." -- FP top of "
    fdepth dup .  #16 min 0	\ limit to 16 items
    ?do  cr i fpick f.  loop
  else
    cr ." Empty FP stack"
  endif
  cr
;


\ *********************
\ Environment Variables
\ *********************

also environment definitions

true		constant	FLOATING
true		constant	FLOATING-EXT
#4096 FPCELL /	constant	FLOATING-STACK

pad 3 cells $FF fill

FPCELL #10 = [if]  $00007FFE  pad 2 cells + !  [then]
FPCELL 8   = [if]  $7FEFFFFF  pad 1 cells + !  [then]
FPCELL 4   = [if]  $7EFFFFFF  pad           !  [then]

pad f@		fconstant	MAX-FLOAT

#fdigits	constant	MAX-FLOAT-DIGITS	( not ANS )

previous definitions


\ *********************************************
\ Patch the floating point pack into the system
\ *********************************************

defer old-number?
  action-of number? to-do old-number?

: fnumber?  	\ c-addr -- 0 | n 1 | d 2 | -2 F: -- r
( *G Behaves like the integer version of NUMBER? except that if )
( ** integer conversion fails, and BASE is decimal, a floating )
( ** point conversion is attempted. If conversion is successful, )
( ** the floating point number is left on the float stack and )
( ** the result code is -2. )
  dup >r  old-number? dup 0= if		\ test for integer first
    base @ #10 = if			\ not integer, floats must be in DECIMAL
      drop  r@ count >float		\ >FLOAT returns well formed flag
      -2 and				\ good float returns -2,
    then
  then
  r> drop
;

assign fnumber? to-do number?
assign #fliteral to-do c#fliteral

: .NSW		\ sw -- ; display NDP status word
  ." NDP SW = " .word space
;

: .FSysPrompt	\ -- ; adds floating point stack depth display
  .SysPrompt
  fdepth if
    ."  F:-" fdepth .
  endif
  ndepth 1 <> if
    ."  N:-" ndepth .
  endif
  sw@ dup $0040 and if
    cr ." NDP Stack Fault: " dup .NSW  fclex
  endif
  dup $001D and if	\ ignore precision and denormals
    cr ." NDP Potential Exception: " dup .NSW  fclex
  endif
  drop
;
assign .FSysPrompt to-do .prompt

: reals		\ -- ; turn FP system on
( *G Enables the floating point package for number conversion. )
  assign fnumber? to-do number?
  assign #fliteral to-do c#fliteral
  assign .FSysPrompt to-do .prompt
;

: integers	\ -- ; turn FP system off
( *G Disables the floating point package for number conversion. )
  [ also system ]
[defined] WinInteger? [if]
  assign wininteger? to-do number?
[else]
  assign integer? to-do number?
[then]
  [ previous ]
  assign (fliteral) to-do c#fliteral
  assign .SysPrompt to-do .prompt
;


\ ********************
\ PFW2.x compatibility
\ ********************

: (f#)		\ addr -- fn ior ; counted string to FP number
  fnumber?
  case
    0  of  0  endof
    1  of  s>f FpSystem  endof
    2  of  d>f FpSystem  endof
  end-case
;

: f#		\ -- f ; or compiles it [ state smart ]
( *G Used in the form "F# <number>", the <number> string is converted	)
( ** and promoted if required to a floating point number. If the system	)
( ** is compiling the float is compiled. If <number> cannot be		)
( ** an error occurs.							)
  bl word (f#) ?undef
  state @
  if  [compile] fliteral  endif
; immediate


\ *****************
\ Debugging support
\ *****************

( *S Debugging facilities )
( *P Debugging floating point code is often difficult, as failures )
( ** can occur because of the necessary approximations involved )
( ** in floating point operations. )

( *P If you set the constant [FPDEBUG] true when compiling HFP387.FTH, )
( ** a debug build will be constructed. The state of the FPU will be )
( ** checked after each word. If a floating point exception )
( ** has been generated, a diagnostic is issued, and the system )
( ** aborts. Set this only when testing, as it slows down the normal )
( ** operation of floating point words. )

( *P The debugger works by intercepting the end of each code definition )
( ** which is finished by FNEXT, rather than the normal NEXT, or RET. )
( ** See the source code in LIB\HFP387.FTH for more details. )

((
NDP status word:
bit	Function
================
15	Busy
14	C3
13..11	TOP
10	C2
9	C1
8	C0
7	Set on Unmasked Exception
6	Set on stack fault (bit 9 set for overflow, clear for underflow)
5	Set on Precision Exception
4	Set on Underflow
3	Set on Overflow
2	Set on Zero Divide
1	Set on Denormalised Operand
0	Set on Invalid Operation
))

[fpdebug] [if]
: .NDP		\ -- ; what to do if exception found
  sw@ dup .NSW
\  dup $20 and if ." Precision " endif
  dup $10 and if ." Underflow " endif
  dup $08 and if ." Overflow " endif
  dup $04 and if ." ZeroDivide " endif
\  dup $02 and if ." Denormalised " endif
  dup $01 and if ." Invalid " endif
  drop
  fclex				\ clear exceptions
  hex .rs  decimal		\ Return stack trace
  finit abort			\ reset stack and restart
;

: (fpcheck)	\ -- ; action of debugger
  sw@ $0040 and if
    cr ." NDP Stack Fault: " .NDP
  endif
  sw@ $001D and if	\ ignore precision and denormals
    cr ." NDP Exception: " .NDP
  endif
;

: +fpcheck	 \ -- ; enable FP checking
( *G Enables the floating point debugger if it has been compiled. )
  assign (fpcheck) to-do fpcheck
;

: -fpcheck	 \ -- ; disable FP checking
( *G Disbles the floating point debugger if it has been compiled. )
  assign noop to-do fpcheck
;
-fpcheck
[then]


\ ======
\ *> ###
\ ======

((
\ test code

decimal

finit rounded

#256 buffer: foo

: t		\ -- ; f: f --
  foo >bcd .
  foo #80 dump
;

: tr		\ --
  cr  #1000 0 do
    cr
    i 5 bounds do
      i s>f  ( 0.1e0 f+ ) fs.  out @ #15 + #-16 and >pos
    loop
  5 +loop
  cr
;

finit rounded
))

