\ DotPsVfx.fth - Dot notation parser for CLASSVFX system

((
Copyright (c) 1999, 2000, 2010, 2011
Construction Computer Software
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691

net: tech-support@mpeforth.com
web: www.mpeforth.com
))


[FORTH]

\ ===========
\ *> classvfx
\ ===========

\ **********************
\ *S Dot notation parser
\ **********************
( *P In order to deal with structures and fields without having to backtrack the )
( ** input stream or the execution order, an additional stage is added to the    )
( ** Forth parser to allow phrases of the forms:                                 )
( *C   inst.field                                                                )
( *C   inst.field.field                                                          )
( *C   type.field                                                                )
( *C   type.field.field                                                          )
( *P to be parsed, where each item is separated by a dot character. The first item  )
( ** must be an instance of a type or a type. If it is an instance, the address is  )
( ** provided, otherwise the base address is assumed to be on the stack. Any items  )
( ** between the first and last item add their offsets to the address, and the last )
( ** item performs the usual operation of the field as defined by an operator. For  )
( ** example:                                                                    )

( *E   type: point:                                                              )
( **     int: x0                                                                 )
( **     int: y0                                                                 )
( **                                                                             )
( **     :m <op1> ... ;m                                                         )
( **     :m <op2> ... ;m                                                         )
( **   end-type                                                                  )
( **                                                                             )
( **   point: Mypoint                                                            )
( **     5 to MyPoint.x0                                                         )

( *P We might define a line as joining two points:                               )
( *E   type: line:                                                               )
( **     point: p1                                                               )
( **     point: p2                                                               )
( **     ...                                                                     )
( **   end-type                                                                  )
( **                                                                             )
( **   line: MyLine                                                              )
( **     5 to MyLine.p2.y0                                                       )

: COMP-1ST-STRUCT  **	\ operator cfa flag --
\ Compile the first part of a dotted phrase.
\ Instance of TYPE: or POINTER: are the only valid cfa's.
   { %operator %cfa %immflag | %pfa %id --- }
   %cfa >body <to> %pfa
   %pfa is-inst?                                          \ instance of type: ??
   IF   \ yes! can be first element in dot parser
        True
   ELSE \ no! test for pointer to instance of type:
        %pfa INST>TYPE INFORTH?                  \ addr at ^type within forth ??
        IF   \ yes! check if addr is an instance of type:
             %pfa INST>TYPE is-type?           \ pointer point to valid type: ??
\             IF  \ yes! must be a pointer
\                  True
\             ELSE \ no! invalid pointer
\                  False
\             THEN
        ELSE \ no! invalid pointer
             False
        THEN
   THEN
   IF
        %operator op# <offsetof> <>
        %operator op# <sizeof>   <> AND
        IF
             %cfa >body TYPE-ID@ ?DUP
             IF
                  <to> %id
             ELSE
                  S" CLASS Internal Software Error" COMP-$ABORT
             THEN
             %id INST:-ID =
             IF
                  OP# <addr> OperatorType !
             ELSE
                  OP# <ptr@> OperatorType !
             THEN
             %cfa execute
        THEN
   ELSE
        %cfa XT>NFA COUNT >ABORT-FUNC$
         S" Use Instance of type only !" COMP-$ABORT
   THEN
;

: COMP-MIDDLE-STRUCT **	( operator cfa flag --- )
\ Compile the middle portions of a dotted phrase.
\ Instance of TYPE: is the only valid cfa.
   { %operator %cfa %immflag -- }
   %cfa >body is-inst?
   IF
        %operator op# <sizeof> <>
        IF
             %operator op# <offsetof> =
             IF
                  %operator OperatorType !
             ELSE
                  op# +offsetof  OperatorType !
             THEN
             %cfa execute
        THEN
   ELSE
        %cfa XT>NFA COUNT >ABORT-FUNC$
        S" Use Instance of type only !" COMP-$ABORT
   THEN
  ;

: COMP-LAST-STRUCT  **	\ operator cfa flag --
\ Compile the last part of a dotted phrase.
   { %operator %cfa %immflag | %pfa --- }
   %cfa >body <to> %pfa
   %pfa is-type? %pfa is-inst? OR            \ type: def or instance of type: ??
   IF
        %operator OperatorType !
        %pfa is-inst?
        IF
             %cfa execute
        ELSE
             %cfa >BODY EX-INST-METHOD-CHAIN
        THEN
   ELSE
        %cfa XT>NFA COUNT >ABORT-FUNC$
        S" Use Instance of type only !" COMP-$ABORT
   THEN
  ;

DEFER dbgSET-TYPEVOC  **	( pfa --- )
ASSIGN DROP TO-DO dbgSET-TYPEVOC
VARIABLE dbgVar **

: SET-TYPEVOC  **	\ xt --
\ set top of search order to TYPE: private wordlist
   { %cfa | %pfa --- }
   %cfa >body <to> %pfa
   %pfa IS-TYPE?                                         \ if a TYPE: definition
   IF   \ yes! type: defintion
        %pfa type-wordlist context !
        #1 dbgVar ! %pfa dbgSET-TYPEVOC
        exit
   THEN
   %pfa IS-INST?                               \ is xt an instance of a type: ??
   IF \ yes! instance of type
        %pfa INST>TYPE type-wordlist context !
        #2 dbgVar ! %pfa dbgSET-TYPEVOC
        exit
   THEN
   %pfa INST>TYPE INFORTH?
   IF   \ yes!
        %pfa INST>TYPE IS-TYPE?                \ pointer point to valid type: ??
        IF  \ yes!
             #3 dbgVar ! %pfa dbgSET-TYPEVOC
             %pfa TYPE-ID@ type-wordlist context !
             exit
        THEN
   THEN
   S" Set Search Order for TYPE: members" >ABORTCAPTION$
   %cfa XT>NFA COUNT >ABORT-FUNC$
   S" not a TYPE: or child of TYPE:"  COMP-$ABORT
  ;

DEFER dbg(comp-struct)	**	( pfa --- )
  ASSIGN DROP TO-DO dbg(comp-struct)

: TYPE-INST-MEMBER?	**	( type inst --- flag )
   { %type %inst | %flag %next --- }
   <zero> %flag
   %type field-link @ <to> %next
   BEGIN
      %next
   WHILE
        %next %inst =
        IF
             <zero> %next
             <set> %flag
        ELSE
             %next field-link @ <to> %next
        THEN
   REPEAT
   %flag
;

: (comp-struct)	**	\ -- flag ; process aaa.bbb.ccc from i/p stream ;\ WJB004
\ The stream is treated as first.middles.last
\ If the operator is ADDR or OFFSETOF, change the last operation to
\ be performed, and then do the first operation.
\ The first operator in the dot sequence may be a colon definition that
\ provides the address for the rest of the sequence.
   { | %operator %immflag %cfa %error %pfa %type -- }
   zBuffer: %zbuf

   <zero> %error
   <zero> %type
   discard-sinline
   BL parse-leading
   pocket count <to> %zbuf
   [CHAR] . word find ?dup                              \ does the word exist ??
   IF
        <to> %immflag                \ yes! get the immediate status of the word
        <to> %cfa                                   \ ...get the cfa of the word
        %immflag 0<>                                                  \ found ??
        IF   \ yes! maybe
             %cfa >BODY <to> %pfa                     \ get the body of the word
             OperatorType @  <to> %operator
            %cfa Set-TypeVoc                     \ Open the wordlist, process it
            %pfa IS-TYPE?                                \ a type: definition ??
            IF   \ yes!
                 %pfa <to> %type
                 #1 dbgVar ! %pfa dbg(comp-struct)
            ELSE \ no! must be instance or pointer ??
                 %pfa IS-INST?                          \ a instance of type: ??
                 IF   \ yes!
                      %pfa INST>TYPE <to> %type
                      #2 dbgVar ! %pfa dbg(comp-struct)
                 ELSE \ no! must be a child of ptr:
                      %pfa TYPE-ID@ <to> %type
                      #3 dbgVar ! %pfa dbg(comp-struct)
                 THEN
                 %operator op# <addr> =
                 IF
                      OP# +offsetof >R
                 ELSE
                      OperatorType @ >R
                 THEN
                 %operator %cfa %immflag COMP-1ST-STRUCT
                 R> <to> %operator
                 <count> %zbuf pocket place
            THEN
            \ Now process the middles with the +OFFSETOF method
            BEGIN
            [CHAR] . WORD FIND
            <to> %immflag                 \ get the immediate status of the word
            <to> %cfa                               \ ...get the cfa of the word
            %immflag 0<>                                              \ found ??
            IF
                 %cfa >BODY <to> %pfa
                 %cfa SET-TYPEVOC               \ open required private wordlist
                 #4 dbgVar ! %pfa dbg(comp-struct)
                 >in @ #tib @ <>                   \ yes! end of input stream ??
                 IF
                      %pfa IS-INST?                     \ a instance of type: ??
                       IF \ yes!
                            %type %pfa TYPE-INST-MEMBER?
                            IF \ yes!
                                 %operator %cfa %immflag comp-middle-struct
                                 %operator op# <offsetof> =
                                 IF
                                      op# +offsetof <to> %operator
                                 THEN
                                 %pfa INST>TYPE <to> %type
                                 #5 dbgVar ! %pfa dbg(comp-struct)
                                 False
                            ELSE
                                 S" Not a valid member!"
                                 COMP-$ABORT
                            THEN
                       ELSE
                            S" Use Instance of type only !"
                            COMP-$ABORT
                       THEN
               ELSE
                    %pfa IS-INST?                        \ a instance of type...
                    %pfa IS-TYPE? OR               \ ...or a type: definition ??
                    IF
                         %pfa IS-INST?                  \ a instance of type: ??
                         IF \ yes!
                              %type %pfa TYPE-INST-MEMBER? 0=
                              IF \ yes!
                              S" Not a valid memeber!"
                              COMP-$ABORT
                              THEN
                         THEN
                         %operator %cfa %immflag comp-last-struct
                         #6 dbgVar ! %pfa dbg(comp-struct)
                         True True              \ return true flag and exit loop
                    ELSE
                         S" Use Instance of type or type!"
                         COMP-$ABORT
                    THEN
               THEN
           ELSE
               S" Use Instance of type or type!"
               COMP-$ABORT
           THEN
         UNTIL
         ELSE
              False
         THEN
    ELSE
       False
    THEN
;

: comp-struct?  **	\ -- flag
\ The stream is treated as first.middles.last
\ If the operator is ADDR or OFFSETOF, change the last operation to
\ be performed, and then do the first operation.
  save-input bl word find >r drop restore-input drop
  r> if 0 exit then
  save-input
  bl word number? dup >r
  0 ?do  drop  loop
  restore-input drop
  r> if 0 exit then
  also
  ['] (comp-struct) CATCH DUP
  IF
       Abort-Code !
  ELSE
       drop Abort-Code OFF
  THEN
  previous
  Abort-Code @                                   \ was there an compile error ??
  IF   \ yes!
       ACTION-OF DISP-ABORT ASSIGN NOOP =                      \ in a console ??
       IF   \ yes!
            Abort-Code @OFF THROW
       ELSE \ no! display ccs system
            DISP-ABORT
            False
       THEN
  THEN
;

: StructHook  **	\ c-addr u -- flag
\ Return true if compiled properly
    pocket place                                           \ copy away from HERE
    save-input n>r
    >in off  blk off
    pocket count to-source -1 to source-id
    comp-struct?
    nr> restore-input drop
;

: +structures	runword	\ --
\ *G Switch on the structure compiler.
  assign StructHook to-do ClassHook
;

: -structures   runword	\ --
\ *G Switch off the structure compiler.
  assign (classhook) to-do ClassHook
;


\ ======
\ *> ###
\ ======

  HIDE-**
